""" Data Formatting.

    Copyright (c) 2009, 2011 The PyroScope Project <pyroscope.project@gmail.com>
"""


import datetime
import logging
import time

from pprint import pformat
from typing import Optional


log = logging.getLogger(__name__)


def human_size(size: float) -> str:
    """Return a human-readable representation of a byte size.

    @param size: Number of bytes as an integer or string.
    @return: String of length 10 with the formatted result.
    """
    if isinstance(size, str):
        size = float(size, 10)

    if size < 0:
        return "-??? bytes"

    if size < 1024:
        return f"{int(size):4d} bytes".lstrip()

    rem = float(size)
    for unit in ("KiB", "MiB", "GiB", "TiB", "PiB"):
        rem /= 1024.0
        if rem < 1024:
            return f"{rem:6.1f} {unit}".lstrip()

    return f"{rem:6.1f} PiB".lstrip()


def iso_datetime(timestamp: Optional[float] = None) -> str:
    """Convert UNIX timestamp to ISO datetime string.

    @param timestamp: UNIX epoch value (default: the current time).
    @return: Timestamp formatted as "YYYY-mm-dd HH:MM:SS".
    """
    if timestamp is None:
        timestamp = time.time()
    return datetime.datetime.fromtimestamp(timestamp).isoformat(" ")[:19]


def iso_datetime_optional(timestamp) -> str:
    """Convert UNIX timestamp to ISO datetime string, or "never".

    @param timestamp: UNIX epoch value.
    @return: Timestamp formatted as "YYYY-mm-dd HH:MM:SS", or "never" for false values.
    """
    if timestamp:
        return iso_datetime(timestamp)
    return "never"


def human_duration(
    time1: float, time2: Optional[float] = None, precision: int = 0, short: bool = False
) -> str:
    """Return a human-readable representation of a time delta.

    @param time1: Relative time value.
    @param time2: Time base (C{None} for now; 0 for a duration in C{time1}).
    @param precision: How many time units to return (0 = all).
    @param short: Use abbreviations, and right-justify the result to always the same length.
    @return: Formatted duration.
    """
    if time2 is None:
        time2 = time.time()

    duration = (time1 or 0) - time2
    direction = (
        " ago" if duration < 0 else ("+now" if short else " from now") if time2 else ""
    )
    duration = abs(duration)
    parts = [
        ("weeks", duration // (7 * 86400)),
        ("days", duration // 86400 % 7),
        ("hours", duration // 3600 % 24),
        ("mins", duration // 60 % 60),
        ("secs", duration % 60),
    ]

    # Kill leading zero parts
    while len(parts) > 1 and parts[0][1] == 0:
        parts = parts[1:]

    # Limit to # of parts given by precision
    if precision:
        parts = parts[:precision]

    numfmt = ("%d", "%d"), ("%4d", "%2d")
    fmt = "%1.1s" if short else " %s"
    sep = " " if short else ", "
    result = (
        sep.join(
            (numfmt[bool(short)][bool(idx)] + fmt)
            % (val, key[:-1] if val == 1 else key)
            for idx, (key, val) in enumerate(parts)
            if val  # or (short and precision)
        )
        + direction
    )

    if not time1:
        result = "never" if time2 else "N/A"

    if precision and short:
        return result.rjust(1 + precision * 4 + (4 if time2 else 0))
    return result


def convert_strings_in_iter(obj):
    """Helper function to nicely format results"""
    if isinstance(obj, bytes):
        obj = obj.decode()
    elif isinstance(obj, dict):
        for k, v in obj.items():
            obj[convert_strings_in_iter(k)] = convert_strings_in_iter(v)
    elif isinstance(obj, list):
        for k, v in enumerate(obj):
            obj[k] = convert_strings_in_iter(v)
    return obj


def rpc_result_to_string(result) -> str:
    """Helper function to nicely format results"""
    result = convert_strings_in_iter(result)

    if isinstance(result, str):
        return result
    if isinstance(result, bytes):
        return result.decode()
    if hasattr(result, "__iter__"):
        return "\n".join(i if isinstance(i, str) else pformat(i) for i in result)
    return repr(result)



""" Command Line Script Support.

    Copyright (c) 2009, 2010 The PyroScope Project <pyroscope.project@gmail.com>
"""


import errno
import logging.config
import sys
import textwrap
import time
import traceback

from argparse import ArgumentParser
from typing import Iterator, List

from pyrosimple import config, error
from pyrosimple.torrent import rtorrent
from pyrosimple.util import pymagic


class ScriptBase:
    """Base class for command line interfaces."""

    # log level for user-visible standard logging
    STD_LOG_LEVEL = logging.INFO

    # argument description for the usage information
    ARGS_HELP = "<log-base>..."

    # additonal stuff appended after the command handler's docstring
    ADDITIONAL_HELP: List[str] = []

    # Can be empty or None in derived classes
    COPYRIGHT = ""

    # Can be made explicit in derived classes (for external tools)
    VERSION = None

    @classmethod
    def setup(cls, _=None):
        """Set up the runtime environment."""
        logging.basicConfig(level=logging.WARNING)

    def __init__(self):
        """Initialize CLI."""
        self.startup = time.time()
        self.LOG = pymagic.get_class_logger(self)
        self.config_dir = ""

        self.args = None
        self.options = None
        self.return_code = 0
        self.engine = None
        self.intermixed_args = False

        # For python 3.7 compatibility
        try:
            import importlib.metadata  # pylint: disable=import-outside-toplevel

            self.__version__ = importlib.metadata.version(  # pylint: disable=no-member
                "pyrosimple"
            )
        except ImportError:
            self.__version__ = "unknown"
        self.version_info = f"{self.__version__} on Python {sys.version.split()[0]}"

        self.parser = ArgumentParser(
            usage="%(prog)s [options] " + self.ARGS_HELP + "\n\n"
            "%(prog)s "
            + self.version_info
            + ("\n" + self.COPYRIGHT if self.COPYRIGHT else "")
            + "\n\n"
            + textwrap.dedent(self.__doc__.rstrip()).lstrip("\n")
            + "\n".join(self.ADDITIONAL_HELP)
            + "\n\nFor more details, see the full documentation at"
            + "\n\n    https://kannibalox.github.io/pyrosimple/",
        )

        self.parser.add_argument(
            "--version", action="version", version="%(prog)s " + self.version_info
        )
        self.parser.add_argument("args", nargs="*")

    def add_bool_option(self, *args, **kwargs):
        """Add a boolean option.

        @keyword help: Option description.
        """
        dest = (
            [o for o in args if o.startswith("--")][0]
            .replace("--", "")
            .replace("-", "_")
        )
        self.parser.add_argument(
            dest=dest, action="store_true", default=False, help=kwargs["help"], *args
        )

    def add_value_option(self, *args, **kwargs):
        """Add a value option.

        @keyword dest: Destination attribute, derived from long option name if not given.
        @keyword action: How to handle the option.
        @keyword help: Option description.
        @keyword default: If given, add this value to the help string.
        """
        kwargs["metavar"] = args[-1]
        if "dest" not in kwargs:
            kwargs["dest"] = (
                [o for o in args if o.startswith("--")][0]
                .replace("--", "")
                .replace("-", "_")
            )
        if "default" in kwargs and kwargs["default"]:
            kwargs["help"] += f" [{kwargs['default']}]"
        if "choices" in kwargs:
            del kwargs["type"]
        self.parser.add_argument(*args[:-1], **kwargs)

    def get_options(self):
        """Get program options."""
        self.parser.add_argument(
            "-q",
            "--quiet",
            "--cron",  # For backwards compatibility
            help="silence warnings",
            dest="log_level",
            action="store_const",
            const=logging.CRITICAL,
        )
        self.parser.add_argument(
            "--debug",
            help="show detailed messages",
            dest="log_level",
            action="store_const",
            const=logging.DEBUG,
        )
        self.parser.add_argument(
            "-v",
            "--verbose",
            help="show additional information",
            dest="log_level",
            action="store_const",
            const=logging.INFO,
        )

        # Template method to add options of derived class
        self.add_options()

        if self.intermixed_args:
            self.options = self.parser.parse_intermixed_args()
        else:
            self.options = self.parser.parse_args()
        self.args = self.options.args

        if self.options.log_level:
            logging.getLogger().setLevel(self.options.log_level)

        self.LOG.debug(
            "Options: %s",
            ", ".join("%s=%r" % i for i in sorted(vars(self.options).items())),
        )

    def fatal(self, msg, exc=None):
        """Exit on a fatal error."""
        if exc is not None:
            self.LOG.fatal("%s (%s)", msg, exc)
            if logging.getLogger().isEnabledFor(logging.DEBUG):
                return  # let the caller re-raise it
        else:
            self.LOG.fatal(msg)
        sys.exit(error.EX_SOFTWARE)

    def run(self):
        """The main program skeleton."""
        log_total = True

        try:
            try:
                # Preparation steps
                self.get_options()

                # Template method with the tool's main loop
                self.mainloop()
            except error.LoggableError:
                print(traceback.format_exc())
                sys.exit(error.EX_SOFTWARE)
            except KeyboardInterrupt:
                print()
                self.LOG.critical("Aborted by CTRL-C!\n")
                sys.exit(error.EX_TEMPFAIL)
            except OSError as exc:
                # [Errno 32] Broken pipe?
                if exc.errno == errno.EPIPE:
                    print(f"\n{exc}, exiting!\n", file=sys.stderr)
                    sys.exit(error.EX_IOERR)
                else:
                    raise
        finally:
            # Shut down
            if log_total and self.options:  ## No time logging on --version and such
                running_time = time.time() - self.startup
                self.LOG.log(
                    self.STD_LOG_LEVEL, "Total time: %.3f seconds.", running_time
                )
            logging.shutdown()

        # Special exit code?
        if self.return_code:
            sys.exit(self.return_code)

    def add_options(self):
        """Add program options."""

    def mainloop(self):
        """The main loop."""
        raise NotImplementedError()


class ScriptBaseWithConfig(ScriptBase):  # pylint: disable=abstract-method
    """CLI tool with configuration support."""

    OPTIONAL_CFG_FILES: List[str] = []

    def add_options(self):
        super().add_options()
        self.parser.add_argument("-U", "--url", help="URL to rtorrent instance")

    def get_options(self):
        """Get program options."""
        super().get_options()
        if self.options.url:
            config.settings["SCGI_URL"] = self.options.url
        config.load_custom_py()
        self.engine = rtorrent.RtorrentEngine()

    def lookup_connection_alias(self, uri: str) -> str:  # pylint: disable=no-self-use
        """Convert a connection alias to the actual URI (if set in the config"""
        if uri in config.settings["CONNECTIONS"]:
            return str(config.settings["CONNECTIONS"][uri])
        return uri

    def multi_connection_lookup(self, uri: str) -> Iterator[str]:
        """Return a list of uris.

        This is separate from lookup_connection_alias due to scripts needing to be written specifically
        to handle this"""
        val = config.settings["CONNECTIONS"].get(uri, [uri])
        if isinstance(val, list):
            for v in val:
                yield self.lookup_connection_alias(v)
        else:
            yield self.lookup_connection_alias(val)


class PromptDecorator:
    """Decorator for interactive commands."""

    def __init__(self, script_obj):
        """Initialize with containing tool object."""
        self.script = script_obj

    def add_options(self):
        """Add program options, must be called in script's addOptions()."""
        # These options need to be conflict-free to the containing
        # script, i.e. avoid short options if possible.
        self.script.add_bool_option(
            "-i",
            "--interactive",
            help="interactive mode (prompt before changing things)",
        )
        self.script.add_bool_option(
            "--yes", help="positively answer all prompts (e.g. --delete --yes)"
        )

    def ask_bool(self, question, default=True):
        """Ask the user for Y)es / N)o / Q)uit.

        If "Q" is entered, this method will exit with RC=3.
        Else, the user's choice is returned.

        Note that the options --non-interactive and --defaults
        also influence the outcome.
        """
        if self.script.options.yes:
            return True
        if self.script.options.dry_run or not self.script.options.interactive:
            return default

        # Let the user decide
        choice = "*"
        while choice not in "YNAQ":
            choice = input(
                "%s? [%s)es, %s)o, a)ll yes, q)uit]: "
                % (
                    question,
                    "yY"[int(default)],
                    "Nn"[int(default)],
                )
            )
            choice = choice[:1].upper() or "NY"[int(default)]

        if choice == "Q":
            self.quit()
        if choice == "A":
            self.script.options.yes = True
            choice = "Y"

        return choice == "Y"

    def quit(self):
        """Exit the program due to user's choices."""
        self.script.LOG.warn("Abort due to user choice!")
        sys.exit(error.EX_TEMPFAIL)



"""Move a torrent to a remote host

Currently uses consistent hashing to determine which remote host
gets the torrent (if multiple are specified)."""
import hashlib
import os
import xmlrpc.client

from pathlib import Path
from time import sleep
from typing import Dict, List

import bencode

import pyrosimple

from pyrosimple import error
from pyrosimple.util import matching, metafile, pymagic, rpc


def nodes_by_hash_weight(meta_id: str, nodes: List[str]) -> Dict[str, int]:
    """Weight nodes by hashing the meta_id"""
    result = {
        n: int.from_bytes(hashlib.sha256(meta_id.encode() + n.encode()).digest(), "big")
        for n in nodes
    }
    return dict(sorted(result.items(), key=lambda item: item[1]))


def get_custom_fields(infohash, proxy):
    """Try using rtorrent-ps commands to list custom keys, otherwise fall back to reading from a session file."""
    if "d.custom.keys" in proxy.system.listMethods():
        custom_fields = {}
        for key in proxy.d.custom.keys(infohash):
            custom_fields[key] = proxy.d.custom(infohash, key)
    else:
        info_file = Path(proxy.session.path(), f"{infohash}.torrent.rtorrent")
        proxy.d.save_full_session(infohash)
        with open(info_file, "rb") as fh:
            custom_fields = bencode.bread(fh)["custom"]
    return custom_fields


class Mover:
    """Move torrent to remote host(s)"""

    def move(
        self,
        infohash,
        remote_proxy,
        fast_resume=True,
        extra_cmds=None,
        keep_basedir=True,
        copy=False,
    ):
        """Moves a torrent to a specific host"""
        if extra_cmds is None:
            extra_cmds = []
        self.LOG.debug(
            "Attempting to %s %s",
            "copy" if copy else "move",
            infohash,
        )
        try:
            remote_proxy.d.hash(infohash)
        except rpc.HashNotFound:
            pass
        else:
            self.LOG.warning("Hash exists remotely")
            return False

        torrent = bencode.bread(
            os.path.join(self.proxy.session.path(), f"{infohash}.torrent")
        )

        if keep_basedir:
            esc_basedir = self.proxy.d.directory_base(infohash).replace('"', '"')
            extra_cmds.insert(0, f'd.directory_base.set="{esc_basedir}"')

        if self.proxy.d.complete(infohash) == 1 and fast_resume:
            self.LOG.debug(
                "Setting fast resume data from %s",
                self.proxy.d.directory_base(infohash),
            )
            metafile.add_fast_resume(torrent, self.proxy.d.directory_base(infohash))

        xml_metafile = xmlrpc.client.Binary(bencode.bencode(torrent))

        if not copy:
            self.proxy.d.stop(infohash)
        self.LOG.debug("Running extra commands on load: %s", extra_cmds)
        remote_proxy.load.raw("", xml_metafile, *extra_cmds)
        for _ in range(0, 5):
            try:
                remote_proxy.d.hash(infohash)
            except rpc.HashNotFound:
                sleep(1)
        # After 5 seconds, let the exception happen
        remote_proxy.d.hash(infohash)

        # Keep custom values
        for k, v in get_custom_fields(infohash, self.proxy).items():
            remote_proxy.d.custom.set(infohash, k, v)
        for key in range(1, 5):
            value = getattr(self.proxy.d, f"custom{key}")(infohash)
            getattr(remote_proxy.d, f"custom{key}.set")(infohash, value)

        if fast_resume:
            remote_proxy.d.start(infohash)
        if not copy:
            self.proxy.d.erase(infohash)
        return True

    def __init__(self, config=None):
        """Initalize torrent mover job"""
        self.config = config or {}
        self.LOG = pymagic.get_class_logger(self)
        self.LOG.debug("Statistics logger created with config %r", self.config)
        self.config.setdefault("dry_run", False)
        self.proxy = None
        self.engine = None

    def run(self):
        """Statistics logger job callback."""
        try:
            self.engine = pyrosimple.connect()
            self.proxy = self.engine.open()
            matcher = matching.create_matcher(self.config["matcher"])
            hosts = self.config["hosts"]
            if not isinstance(hosts, list):
                hosts = [hosts]
            for i in self.engine.view("default", matcher):
                for host in nodes_by_hash_weight(i.hash + i.alias, hosts):
                    rproxy = rpc.RTorrentProxy(host)
                    metahash = i.hash
                    if self.config["dry_run"]:
                        self.LOG.info(
                            "Would move %s to %s", metahash, rproxy.system.hostname()
                        )
                        break
                    if self.move(metahash, rproxy):
                        self.LOG.info(
                            "Moved %s to %s", metahash, rproxy.system.hostname()
                        )
                        break
        except (error.LoggableError, *rpc.ERRORS) as exc:
            self.LOG.warning(str(exc))



""" Metafile Creator.

    Copyright (c) 2009, 2010, 2011 The PyroScope Project <pyroscope.project@gmail.com>
"""


import hashlib
import logging
import os
import re

from urllib.parse import parse_qs

import bencode

from pyrosimple import config
from pyrosimple.scripts.base import ScriptBase, ScriptBaseWithConfig
from pyrosimple.util import metafile


class MetafileCreator(ScriptBaseWithConfig):
    """
    Create a bittorrent metafile.

    If passed a magnet URI as the only argument, a metafile is created
    in the directory specified via the configuration value 'magnet_watch',
    loadable by rTorrent. Which means you can register 'mktor' as a magnet:
    URL handler in Firefox.
    """

    # argument description for the usage information
    ARGS_HELP = "<dir-or-file> <tracker-url-or-alias>... | <magnet-uri>"

    def add_options(self):
        """Add program options."""
        super().add_options()

        self.add_bool_option("-p", "--private", help="disallow DHT and PEX")
        self.add_bool_option("--no-date", help="leave out creation date")
        self.add_value_option(
            "-o",
            "--output-filename",
            "PATH",
            help="optional file name (or target directory) for the metafile",
        )
        self.add_value_option(
            "-r",
            "--root-name",
            "NAME",
            help="optional root name (default is basename of the data path)",
        )
        self.add_value_option(
            "-m",
            "--magnet-watch",
            "NAME",
            help="path to place .meta files from magnet links",
        )
        self.add_value_option(
            "-x",
            "--exclude",
            "PATTERN",
            action="append",
            default=[],
            help="exclude files matching a glob pattern from hashing; can be specified multiple times",
        )
        self.add_value_option(
            "--comment", "TEXT", help="optional human-readable comment"
        )
        self.add_value_option(
            "-s",
            "--set",
            "KEY=VAL",
            action="append",
            default=[],
            help="set a specific key to the given value; omit the '=' to delete a key; can be specified multiple times",
        )
        self.add_bool_option(
            "-H",
            "--hashed",
            "--fast-resume",
            help="create second metafile containing libtorrent fast-resume information",
        )

    # TODO: Optionally pass torrent directly to rTorrent (--load / --start)
    # TODO: Optionally limit disk I/O bandwidth used (incl. a config default!)
    # TODO: Set "encoding" correctly
    # TODO: Support multi-tracker extension ("announce-list" field)
    # TODO: DHT "nodes" field?! [[str IP, int port], ...]
    # TODO: Web-seeding http://www.getright.com/seedtorrent.html
    #       field 'url-list': ['http://...'] on top-level

    def make_magnet_meta(self, magnet_uri):
        """Create a magnet-uri torrent."""

        if magnet_uri.startswith("magnet:"):
            magnet_uri = magnet_uri[7:]
        meta = {"magnet-uri": "magnet:" + magnet_uri}
        magnet_params = parse_qs(magnet_uri.lstrip("?"))

        meta_name = magnet_params.get("xt", [hashlib.sha1(magnet_uri).hexdigest()])[0]
        if "dn" in magnet_params:
            meta_name = f"{magnet_params['dn'][0]}-{meta_name}"
        meta_name = (
            re.sub(r"[^-_,a-zA-Z0-9]+", ".", meta_name)
            .strip(".")
            .replace("urn.btih.", "")
        )

        if not self.options.magnet_watch:
            self.fatal("You MUST set the '--magnet-watch' config option!")
        meta_path = os.path.join(
            self.options.magnet_watch, f"magnet-{meta_name}.torrent"
        )
        self.LOG.debug("Writing magnet-uri metafile %r...", meta_path)

        try:
            bencode.bwrite(meta_path, meta)
        except OSError as exc:
            self.fatal("Error writing magnet-uri metafile %r (%s)", (meta_path, exc))
            raise

    def mainloop(self):
        """The main loop."""
        if len(self.args) == 1 and "=urn:btih:" in self.args[0]:
            # Handle magnet link
            self.make_magnet_meta(self.args[0])
            return

        if not self.args:
            self.parser.print_help()
            self.parser.exit()
        elif len(self.args) < 2:
            self.parser.error(
                "Expected a path and at least one announce URL, got: %s"
                % (" ".join(self.args),)
            )

        # Create and configure metafile factory
        datapath = self.args[0].rstrip(os.sep)
        metapath = datapath
        if self.options.output_filename:
            metapath = self.options.output_filename
            if os.path.isdir(metapath):
                metapath = os.path.join(metapath, os.path.basename(datapath))
        if not metapath.endswith(".torrent"):
            metapath += ".torrent"
        torrent = metafile.Metafile(metapath)
        torrent.ignore.extend(self.options.exclude)

        def callback(meta):
            "Callback to set label and resume data."
            meta["info"]["source"] = config.map_announce2alias(
                meta["announce"] or meta["announce-list"][0]
            )
            # Set specific keys?
            metafile.assign_fields(meta, self.options.set)

        # Create and write the metafile(s)
        # TODO: make it work better with multiple trackers (hash only once), also create fast-resume file for each tracker
        meta = torrent.create(
            datapath,
            self.args[1:],
            progress=metafile.console_progress()
            if logging.getLogger().isEnabledFor(logging.WARNING)
            else None,
            root_name=self.options.root_name,
            private=self.options.private,
            no_date=self.options.no_date,
            comment=self.options.comment,
            created_by="PyroSimple",
            callback=callback,
        )

        # Create second metafile with fast-resume?
        if self.options.hashed:
            try:
                metafile.add_fast_resume(meta, datapath)
            except OSError as exc:
                self.fatal(f"Error making fast-resume data ({exc})")
                raise

            hashed_path = re.sub(r"\.torrent$", "", metapath) + "-resume.torrent"
            self.LOG.info("Writing fast-resume metafile %r...", hashed_path)
            try:
                bencode.bwrite(hashed_path, meta)
            except OSError as exc:
                self.fatal(
                    f"Error writing fast-resume metafile {hashed_path!r} ({exc})"
                )
                raise


def run():  # pragma: no cover
    """The entry point."""
    ScriptBase.setup()
    MetafileCreator().run()


if __name__ == "__main__":
    run()



""" rTorrent Queue Manager.

    Copyright (c) 2012 The PyroScope Project <pyroscope.project@gmail.com>
"""

import time

from pyrosimple import error
from pyrosimple.job.base import MatchableJob
from pyrosimple.util import matching, rpc


# pylint: disable=abstract-method
class QueueManager(MatchableJob):
    """rTorrent queue manager implementation."""

    def __init__(self, config=None):
        """Set up queue manager."""
        self.config = config or {}
        print(self.config)
        super().__init__(config)
        self.proxy = None
        self.last_start = 0

        bool_param = lambda key, default: matching.truth(self.config.get(key, default))
        self.config.setdefault("viewname", "main")
        self.config.setdefault("downloading_min", 0)
        self.config.setdefault("downloading_max", 20)
        self.config.setdefault("max_downloading_traffic", 0)
        self.config["log_to_client"] = bool_param("log_to_client", True)
        self.log.info(
            "Startable matcher is: %s",
            self.config["matcher"],
        )
        self.config["downloading"] = matching.create_matcher(
            self.config.get("downloading", "is_active=1 is_complete=0")
        )
        self.log.info(
            "Downloading matcher is: %s",
            self.config["downloading"],
        )

    def _start(self, items):
        """Start some items if conditions are met."""
        # TODO: Filter by a custom date field, for scheduled downloads starting at a certain time
        # or after a given delay

        # TODO: Don't start anything more if download BW is used >= config threshold in %

        # Check if anything more is ready to start downloading
        startable = [i for i in items if self.matcher.match(i)]
        if not startable:
            self.log.debug(
                "Checked %d item(s), none startable according to %s",
                len(items),
                self.config["matcher"],
            )
            return

        # Check intermission delay
        intermission = self.config.get("intermission", 120)
        now = time.time()
        if now < self.last_start:
            # compensate for summer time and other oddities
            self.last_start = now
        delayed = int(self.last_start + intermission - now)
        if delayed > 0:
            self.log.debug(
                "Delaying start of %d item(s),"
                " due to %ds intermission with %ds left",
                len(startable),
                intermission,
                delayed,
            )
            return

        # Stick to "start_at_once" parameter, unless "downloading_min" is violated
        downloading = [i for i in items if self.config["downloading"].match(i)]
        start_now = max(
            self.config.get("start_at_once", 1),
            self.config["downloading_min"] - len(downloading),
        )
        start_now = min(start_now, len(startable))

        if self.config["max_downloading_traffic"] > 0:
            down_traffic = sum(i.down for i in downloading)
            self.log.debug("%d downloading, down %d", len(downloading), down_traffic)
            if down_traffic > int(self.config["max_downloading_traffic"]):
                self.log.debug("Max download traffic reaching, skipping start")
                return

        # Start eligible items
        for idx, item in enumerate(startable.sort(key=self.sort_key)):
            # Check if we reached 'start_now' in this run
            if idx >= start_now:
                self.log.debug(
                    "Only starting %d item(s) in this run, %d more could be downloading",
                    start_now,
                    len(startable) - idx,
                )
                break

            # TODO: Prevent start of more torrents that can fit on the drive (taking "off" files into account)
            # (restarts items that were stopped due to the "low_diskspace" schedule, and also avoids triggering it at all)

            # Only check the other conditions when we have `downloading_min` covered
            if len(downloading) < self.config["downloading_min"]:
                self.log.debug(
                    "Catching up from %d to a minimum of %d downloading item(s)",
                    len(downloading),
                    self.config["downloading_min"],
                )
            else:
                # Limit to the given maximum of downloading items
                if len(downloading) >= self.config["downloading_max"]:
                    self.log.debug(
                        "Already downloading %d item(s) out of %d max, %d more could be downloading",
                        len(downloading),
                        self.config["downloading_max"],
                        len(startable) - idx,
                    )
                    break

            # If we made it here, start it!
            self.last_start = now
            downloading.append(item)
            if self.config["dry_run"]:
                self.log.info(
                    "Would start %s '%s'",
                    item.hash,
                    item.name,
                )
            else:
                self.log.info(
                    "Starting %s '%s'",
                    item.hash,
                    item.name,
                )
                item.start()
                if self.config["log_to_client"]:
                    self.engine.open().log(
                        rpc.NOHASH,
                        f"{self.__class__.__name__}: Started '{item.name}' {item.alias}",
                    )

    def run(self):
        """Queue manager job callback."""
        try:
            # Get items from 'main' view
            items = list(self.engine.view(self.config["view"]))

            items.sort(key=self.sort_key)

            # Handle found items
            self._start(items)
        except (error.LoggableError, *rpc.ERRORS) as exc:
            self.log.warning(str(exc))



""" RTorrent client proxy.

    Copyright (c) 2011 The PyroScope Project <pyroscope.project@gmail.com>
"""


import functools
import json
import logging
import random
import urllib

from xmlrpc import client as xmlrpclib

from pyrosimple.io import scgi


logger = logging.getLogger(__name__)

NOHASH = (
    ""  # use named constant to make new-syntax commands with no hash easily searchable
)


CACHE_METHOD = {
    "d.chunk_size",
    "d.is_private",
    "d.name",
    "d.size_bytes",
    "d.size_chunks",
    "d.size_files",
    "system.api_version",
    "system.client_version",
    "system.library_version",
}


class XmlRpcError(xmlrpclib.Fault):
    """Base class for XMLRPC protocol errors."""

    def __init__(self, msg, *args):
        super().__init__(self, msg, *args)
        self.message = msg.format(*args)
        self.faultString = self.message
        self.faultCode = -500

    def __str__(self):
        return self.message


class HashNotFound(XmlRpcError):
    """Non-existing or disappeared hash."""

    def __init__(self, msg, *args):
        super().__init__(msg, *args)
        self.faultCode = -404


ERRORS = (XmlRpcError,) + scgi.ERRORS


class RTorrentProxy(xmlrpclib.ServerProxy):
    # pylint: disable=super-init-not-called
    """Proxy to rTorrent's RPC interface.

    Method calls are built from attribute accesses, i.e. you can do
    something like C{proxy.system.client_version()}.

    All methods from ServerProxy are being overridden due to the combination
    of self.__var name mangling and the __call__/__getattr__ magic.
    """

    def __init__(
        self,
        uri,
        transport=None,
        encoding=None,
        verbose=False,
        allow_none=False,
        use_datetime=False,
        use_builtin_types=False,
        *,
        headers=(),
        context=None,
    ):
        # Get the url
        parsed_url = urllib.parse.urlsplit(uri)
        queries = urllib.parse.parse_qs(parsed_url.query)
        if parsed_url.scheme not in ("http", "https", "scgi", "scgi+ssh", "scgi+unix"):
            raise OSError(
                f"unsupported RPC scheme '{parsed_url.scheme}' in url '{uri}'"
            )
        # Config the connection details
        self.__rpc_codec = queries.get("rpc", ["xml"])[0]
        self.__uri = uri
        self.__host = parsed_url.netloc
        self.__handler = urllib.parse.urlunsplit(["", "", *parsed_url[2:]])
        if not self.__handler and parsed_url.scheme in ("http", "https"):
            self.__handler = "/RPC2"

        if transport is None:
            if self.__rpc_codec == "json":
                codec = json
                headers = [("CONTENT_TYPE", "application/json")]
            elif self.__rpc_codec == "xml":
                codec = xmlrpclib
            handler = scgi.transport_from_url(uri)
            transport = handler(
                uri=uri,
                use_datetime=use_datetime,
                use_builtin_types=use_builtin_types,
                codec=codec,
                headers=headers,
            )
        self.__transport = transport
        self.__encoding = encoding or "utf-8"
        self.__verbose = verbose
        self.__allow_none = allow_none

    def __close(self):
        self.__transport.close()

    def __request_xml(self, methodname, params):
        # Verbatim from parent method
        request = xmlrpclib.dumps(
            params,
            methodname,
            encoding=self.__encoding,
            allow_none=self.__allow_none,
        ).encode(self.__encoding, "xmlcharrefreplace")
        if self.__verbose:
            print("req: ", request)

        response = self.__transport.request(
            self.__host, self.__handler, request, verbose=self.__verbose
        )

        if len(response) == 1:
            response = response[0]

        return response

    def __request_json(self, methodname, params):
        if not params:
            params = [""]

        # Gross hack, see about getting call implemented upstream
        if methodname == "system.multicall":
            results = []
            for i in params[0]:
                results.append([self.__request_json(i["methodName"], i["params"])])
            return results

        # This random ID feels silly but there's not much need for anything better at the moment.
        rpc_id = random.randint(0, 100)
        request = json.dumps(
            {
                "params": params,
                "method": methodname,
                "jsonrpc": "2.0",
                "id": rpc_id,
            }
        ).encode(self.__encoding, "xmlcharrefreplace")
        if self.__verbose:
            print("req: ", request)

        response = self.__transport.request(
            self.__host,
            self.__handler,
            request,
            verbose=self.__verbose,
        )

        if response["id"] != rpc_id:
            raise ValueError(
                f"RPC IDs do not match: sent={rpc_id} received={response['id']}"
            )
        if "error" in response:
            raise ValueError(f"Received error: {response['error']}")
        return response["result"]

    def __request(self, methodname, params):
        """Determines whether or not a request should be cached,
        then passes it to the appropriate method"""
        if methodname in CACHE_METHOD:
            return self.__cached_request(methodname, params)
        return self.__request_switch(methodname, params)

    @functools.lru_cache(maxsize=32)
    def __cached_request(self, methodname, params):
        """Simpled cache pass-through method to more easily take advantage of functools.lru_cache"""
        return self.__request_switch(methodname, params)

    def __request_switch(self, methodname, params):
        """Determines whether the request should go through
        xml or json and calls the appropriate method."""
        logger.debug("method '%s', params %s", methodname, params)
        try:
            if self.__rpc_codec == "xml":
                return self.__request_xml(methodname, params)
            if self.__rpc_codec == "json":
                return self.__request_json(methodname, params)
        except xmlrpclib.Fault as exc:
            if exc.faultString == "Could not find info-hash.":
                raise HashNotFound(  # pylint: disable=raise-missing-from
                    exc.faultString
                )
            raise exc
        raise ValueError(f"Invalid RPC protocol '{self.__rpc_codec}'")

    def __repr__(self):
        return f"<{self.__class__.__name__} via {self.__rpc_codec} for {self.__uri}>"

    def __getattr__(self, name):
        # magic method dispatcher
        # Hardcode the most useful alias
        if name == "log":
            name = "print"
        return xmlrpclib._Method(self.__request, name)

    # note: to call a remote object with a non-standard name, use
    # result getattr(server, "strange-python-name")(args)

    def __call__(self, attr):
        """A workaround to get special attributes on the ServerProxy
        without interfering with the magic __getattr__
        """
        if attr == "close":
            return self.__close
        if attr == "transport":
            return self.__transport
        raise AttributeError(f"Attribute {attr} not found")

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.__close()



""" rTorrent Item Filter Jobs.

    Copyright (c) 2012 The PyroScope Project <pyroscope.project@gmail.com>
"""


from pyrosimple import error
from pyrosimple.util import pymagic, rpc


class FilterJobBase:
    """Base class for filter rule jobs."""

    def __init__(self, filter_config=None):
        """Set up filter config."""
        self.config = filter_config or {}
        self.LOG = pymagic.get_class_logger(self)
        self.LOG.debug(
            "%s created with config %r", self.__class__.__name__, self.config
        )

    def run(self):
        """Filter job callback."""

        try:
            # TODO: select view into items
            items = []
            self.run_filter(items)
        except (error.LoggableError, rpc.ERRORS) as exc:
            self.LOG.warning(str(exc))

    def run_filter(self, items):
        """Perform job on filtered items."""
        raise NotImplementedError()


class ActionRule(FilterJobBase):
    """Perform an action on selected items."""

    def run_filter(self, items):
        """Perform configured action on filtered items."""
        # TODO: what actions? rpc, delete, cull, stop, etc. for sure.


class TorrentMirror(FilterJobBase):
    """Mirror selected items via a specified tracker."""

    def run_filter(self, items):
        """Load filtered items into remote client via tracker / watchdir."""
        # TODO: config is tracker_url, tracker_upload, watch_dir
        # create clones of item's metafile, write to watch_dir, and upload
        # to tracker_upload (support file: at first, for a local bttrack);
        # also, already mirrored items have to be marked somehow



""" rTorrent Proxy.

    Copyright (c) 2009, 2010, 2011 The PyroScope Project <pyroscope.project@gmail.com>
"""


import errno
import fnmatch
import operator
import os
import shlex
import time

from collections import defaultdict
from functools import partial
from pathlib import Path
from typing import Callable, Dict, List, Optional, Set, Union
from xmlrpc import client as xmlrpclib

from pyrosimple import config, error
from pyrosimple.torrent import engine
from pyrosimple.util import fmt, matching, pymagic, rpc, traits
from pyrosimple.util.cache import ExpiringCache
from pyrosimple.util.parts import Bunch


class CommaLexer(shlex.shlex):
    """Helper to split argument lists."""

    def __init__(self, text: str):
        shlex.shlex.__init__(self, text, None, True)
        self.whitespace += ","
        self.whitespace_split = True
        self.commenters = ""


class RtorrentItem(engine.TorrentProxy):
    """A single download item."""

    def __init__(self, engine_, fields, rpc_fields: Optional[Dict] = None):
        """Initialize download item."""
        super().__init__()
        self._engine = engine_
        self._fields = ExpiringCache(
            static_keys=engine.FieldDefinition.CONSTANT_FIELDS
        )  # Acts a cache for the item
        self._fields.update(dict(fields))
        self._rpc_cache = ExpiringCache(
            static_keys={
                "d.name",
                "d.size_bytes",
                "d.size_chunks",
            }
        )
        if rpc_fields is not None:
            self._rpc_cache.update(rpc_fields)
        if "hash" not in fields:
            self._fields["hash"] = self.rpc_call("d.hash")

    def _make_it_so(
        self, command: str, calls: List[str], *args, observer: Optional[Callable] = None
    ):
        """Perform some error-checked RPC calls."""
        args = (self._fields["hash"],) + args
        try:
            self._engine.LOG.debug(
                "%s%s torrent %s",
                command[0].upper(),
                command[1:],
                self._fields["hash"],
            )
            for call in calls:
                namespace = self._engine.rpc
                result = getattr(namespace, call.lstrip(":"))(*args)
                if observer is not None:
                    observer(result)
        except rpc.ERRORS as exc:
            raise error.EngineError(
                f"While {command} torrent {self._fields['hash']}: {exc}"
            )

    def _get_files(self, attrs: Optional[List[str]] = None):
        """Get a list of all files in this download; each entry has the
        attributes C{path} (relative to root), C{size} (in bytes),
        C{mtime}, C{prio} (0=off, 1=normal, 2=high), C{created},
        and C{opened}.

        @param attrs: Optional list of additional attributes to fetch.
        """
        try:
            # Get info for all files
            f_multicall = self._engine.rpc.f.multicall
            f_params = [
                self._fields["hash"],
                0,
                "f.path=",
                "f.size_bytes=",
                "f.last_touched=",
                "f.priority=",
                "f.is_created=",
                "f.is_open=",
            ]
            for attr in attrs or []:
                f_params.append(f"f.{attr}=")
            rpc_result = f_multicall(*tuple(f_params))
        except rpc.ERRORS as exc:
            raise error.EngineError(
                f"While getting files for torrent #{self._fields['hash']}: {exc}"
            )
        else:
            # self._engine.LOG.debug("files result: %r" % rpc_result)

            # Return results
            result = [
                Bunch(
                    path=i[0],
                    size=i[1],
                    mtime=i[2] / 1000000.0,
                    prio=i[3],
                    created=i[4],
                    opened=i[5],
                )
                for i in rpc_result
            ]

            if attrs:
                for idx, attr in enumerate(attrs):
                    for item, rpc_item in zip(result, rpc_result):
                        item[attr] = rpc_item[6 + idx]

            return result

    def memoize(self, name: str, getter: Callable, *args, **kwargs):
        """Cache a stable expensive-to-get item value for later (optimized) retrieval."""
        field = "custom_memo_" + name
        cached = self.rpc_call("d.custom", ["memo_" + name])
        if cached:
            value = cached
        else:
            value = getter(*args, **kwargs)
            self._make_it_so(
                f"caching {name}={value!r} for",
                ["d.custom.set"],
                field[7:],
                value,
            )
            self._fields[field] = value
        return value

    def _get_kind(self, limit):
        """Get a set of dominant file types. The files must contribute
        at least C{limit}% to the item's total size.
        """
        histo = self.fetch("custom_kind")

        if histo:
            # Parse histogram from cached field
            histo = [i.split("%_") for i in str(histo).split()]
            histo = [(int(val, 10), ext) for val, ext in histo]
            ##self._engine.LOG.debug("~~~~~~~~~~ cached histo = %r" % histo)
        else:
            # Get filetypes
            histo = traits.get_filetypes(
                self.fetch("files"),
                path=operator.attrgetter("path"),
                size=operator.attrgetter("size"),
            )

            # Set custom cache field with value formatted like "80%_flac 20%_jpg" (sorted by percentage)
            histo_str = " ".join(("%d%%_%s" % i).replace(" ", "_") for i in histo)
            self._make_it_so(
                f"setting kind cache {histo_str!r} on",
                ["d.custom.set"],
                "kind",
                histo_str,
            )
            self._fields["custom_kind"] = histo_str

        # Return all non-empty extensions that make up at least <limit>% of total size
        return {ext for val, ext in histo if ext and val >= limit}

    def as_dict(self):
        """Return known fields."""
        return self._fields.copy()

    def rpc_call(self, method: str, args: Optional[List] = None, cache: bool = True):
        """Directly call rpc for item-specific information"""
        cache_key = method
        if args:
            cache_key += "=" + ",".join([str(a) for a in args])
        val = self._rpc_cache.get(cache_key, None)
        if cache and val is not None:
            return val
        if args is None:
            args = []
        getter = getattr(self._engine.rpc, method)
        val = getter(self._fields["hash"], *args)
        self._rpc_cache[cache_key] = val
        return val

    def fetch(self, name: str, cache: bool = True):
        """Get a field on demand. By 'on demand', this means that the field may possibly be created
        if it does not already exists (e.g. custom fields). It also allows directly controlling if the _fields cache
        should be used"""
        if cache:
            try:
                return self._fields[name]
            except KeyError:
                pass
        if isinstance(name, int):
            name = "custom_%d" % name
        if name.startswith("kind_") and name[5:].isdigit():
            val = self._get_kind(int(name[5:], 10))
        elif name.startswith("custom_"):
            key = name[7:]
            try:
                if len(key) == 1 and key in "12345":
                    val = getattr(self._engine.rpc.d, "custom" + key)(
                        self._fields["hash"]
                    )
                else:
                    val = self._engine.rpc.d.custom(self._fields["hash"], key)
            except rpc.ERRORS as exc:
                raise error.EngineError(f"While accessing field {name!r}: {exc}")
        else:
            val = getattr(self, name)

        self._fields[name] = val

        return val

    def datapath(self) -> Path:
        """Get an item's data path."""
        path = self.rpc_call("d.directory")
        if path and not self.rpc_call("d.is_multi_file"):
            path = os.path.join(path, self.rpc_call("d.name"))
        path = os.path.expanduser(path)
        if self.rpc_call("d.is_multi_file"):
            return Path(path)
        return Path(path)

    def announce_urls(self, default=[]):  # pylint: disable=dangerous-default-value
        """Get a list of all announce URLs.
        Returns `default` if no trackers are found at all.
        """
        try:
            response = self.rpc_call("t.multicall", ["", "t.url=", "t.is_enabled="])
        except rpc.ERRORS as exc:
            raise error.EngineError(
                f"While getting announce URLs for #{self._fields['hash']}: {exc}"
            )

        if response:
            return [i[0] for i in response if i[1]]
        return default

    def start(self):
        """(Re-)start downloading or seeding."""
        self._make_it_so("starting", ["d.open", "d.start"])

    def stop(self):
        """Stop and close download."""
        self._make_it_so("stopping", ["d.stop", "d.close"])

    def ignore(self, flag: int):
        """Set ignore status."""
        self._make_it_so("setting ignore status for", ["d.ignore_commands.set"], flag)
        self._fields["is_ignored"] = flag

    def set_prio(self, prio: int):
        """Set priority (0-3)."""
        self._make_it_so(
            "setting priority for", ["d.priority.set"], max(0, min(int(prio), 3))
        )

    def tag(self, tags: str):
        """Add or remove tags."""
        # Get tag list and add/remove given tags
        tags = tags.lower()
        previous = self.fetch("tagged")
        tagset = previous.copy()
        for tag in tags.replace(",", " ").split():
            if tag.startswith("-"):
                tagset.discard(tag[1:])
            elif tag.startswith("+"):
                tagset.add(tag[1:])
            else:
                tagset.add(tag)

        # Write back new tagset, if changed
        tagset.discard("")
        if tagset != previous:
            tagset = " ".join(sorted(tagset))
            self._make_it_so(
                f"setting tags {tagset!r} on", ["d.custom.set"], "tags", tagset
            )
            self._fields["custom_tags"] = tagset

    def set_throttle(self, name: str):
        """Assign to throttle group."""
        if name.lower() == "null":
            name = "NULL"
        if name.lower() == "none":
            name = ""

        if name not in self._engine.known_throttle_names:
            if self._engine.rpc.throttle.up.max(rpc.NOHASH, name) == -1:
                if self._engine.rpc.throttle.down.max(rpc.NOHASH, name) == -1:
                    raise error.UserError(f"Unknown throttle name '{name}'")
            self._engine.known_throttle_names.add(name)

        if (name or "NONE") == self.fetch("throttle"):
            self._engine.LOG.debug(
                "Keeping throttle %r on torrent #%s",
                self.fetch("throttle"),
                self._fields["hash"],
            )
            return

        active = self.fetch("is_active")
        if active:
            self._engine.LOG.debug(
                "Torrent #%s stopped for throttling", self._fields["hash"]
            )
            self.stop()
        self._make_it_so(f"setting throttle {name!r} on", ["d.throttle_name.set"], name)
        if active:
            self._engine.LOG.debug(
                "Torrent #%s restarted after throttling",
                self._fields["hash"],
            )
            self.start()

    def set_custom(self, key: str, value: Optional[str] = None):
        """Set a custom value. C{key} might have the form "key=value" when value is C{None}."""
        # Split combined key/value
        if value is None:
            try:
                key, value = key.split("=", 1)
            except (ValueError, TypeError) as exc:
                raise error.UserError(
                    "Bad custom field assignment %r, probably missing a '=' (%s)"
                    % (key, exc)
                )

        # Check identifier rules
        args: List[str]
        if not key:
            raise error.UserError("Custom field name cannot be empty!")
        if len(key) == 1 and key in "12345":
            method, args = "d.custom" + key + ".set", [value]
        elif not (key[0].isalpha() and key.replace("_", "").isalnum()):
            raise error.UserError(
                f"Bad custom field name {key!r} (must only contain a-z, A-Z, 0-9 and _)"
            )
        else:
            method, args = "d.custom.set", [key, value]

        # Make the assignment
        self._make_it_so(f"setting custom_{key} = {value!r} on", [method], *args)
        self._fields["custom_" + key] = value

    def hash_check(self):
        """Hash check a download."""
        self._make_it_so("hash-checking", ["d.check_hash"])

    def __print_result(self, data, method=None, args=None):
        "Helper to print RPC call results"
        args_list = ""
        if args:
            args_list = '"' + '","'.join(args) + '"'
        print(f"{self._fields['hash']}\t{data}\t{method.lstrip(':')}={args_list}")

    def execute(self, commands):
        """Execute RPC command(s)."""
        try:
            commands = [i.strip() for i in commands.split(" ; ")]
        except (TypeError, AttributeError):
            pass  # assume an iterable

        for command in commands:
            try:
                method, args = command.split("=", 1)
                args = tuple(CommaLexer(args))
            except (ValueError, TypeError) as exc:
                raise error.UserError(
                    f"Bad command {command!r}, probably missing a '=' ({exc})"
                )

            observer = None
            if method.startswith(">"):
                observer = partial(self.__print_result, args=args, method=method)
            method = method.lstrip(">")
            self._make_it_so("executing command on", [method], *args, observer=observer)

    def delete(self):
        """Remove torrent from client."""
        self.stop()
        if self.fetch("metafile"):
            self._make_it_so("removing metafile of", ["d.delete_tied"])
        self._make_it_so("erasing", ["d.erase"])

    # TODO: def set_files_priority(self, pattern, prio)
    # Set priority of selected files
    # NOTE: need to call d.update_priorities after f.priority.set!

    def purge(self):
        """Delete PARTIAL data files and remove torrent from client."""

        def partial_file(item):
            "Filter out partial files"
            return item.completed_chunks < item.size_chunks

        self.cull(file_filter=partial_file, attrs=["completed_chunks", "size_chunks"])

    def cull(self, file_filter: Optional[Callable] = None, attrs: List[str] = None):
        """Delete ALL data files and remove torrent from client.

        @param file_filter: Optional callable for selecting a subset of all files.
            The callable gets a file item as described for RtorrentItem._get_files
            and must return True for items eligible for deletion.
        @param attrs: Optional list of additional attributes to fetch for a filter.
        """
        dry_run = 0  # set to 1 for testing

        def remove_with_links(path):
            "Remove a path including any symlink chains leading to it."
            rm_paths = []
            while os.path.islink(path):
                target = os.readlink(path)
                rm_paths.append(path)
                path = target

            if os.path.exists(path):
                rm_paths.append(path)
            else:
                self._engine.LOG.debug(
                    "Real path '%s' doesn't exist,"
                    " but %d symlink(s) leading to it will be deleted..."
                    % (path, len(rm_paths))
                )

            # Remove the link chain, starting at the real path
            # (this prevents losing the chain when there's permission problems)
            for rm_path in reversed(rm_paths):
                is_dir = os.path.isdir(rm_path) and not os.path.islink(rm_path)
                self._engine.LOG.debug(
                    "Deleting '%s%s'", rm_path, "/" if is_dir else ""
                )
                if not dry_run:
                    try:
                        (os.rmdir if is_dir else os.remove)(rm_path)
                    except OSError as exc:
                        if exc.errno == errno.ENOENT:
                            # Seems this disappeared somehow inbetween (race condition)
                            self._engine.LOG.info(
                                "Path '%s%s' disappeared before it could be deleted",
                                rm_path,
                                "/" if is_dir else "",
                            )
                        else:
                            raise

            return rm_paths

        # Assemble doomed files and directories
        files, dirs = set(), set()
        base_path = os.path.expanduser(self.fetch("directory"))
        item_files = list(self._get_files(attrs=attrs))

        if not self.fetch("directory"):
            raise error.EngineError(
                "Directory for item #%s is empty,"
                " you might want to add a filter 'directory=!'"
                % (self._fields["hash"],)
            )
        if not os.path.isabs(base_path):
            raise error.EngineError(
                "Directory '%s' for item #%s is not absolute, which is a bad idea;"
                " fix your .rtorrent.rc, and use 'directory.default.set = /...'"
                % (
                    self.fetch("directory"),
                    self._fields["hash"],
                )
            )
        if self.fetch("is_multi_file") and os.path.isdir(self.fetch("directory")):
            dirs.add(self.fetch("directory"))

        for item_file in item_files:
            if file_filter and not file_filter(item_file):
                continue
            path = os.path.join(base_path, item_file.path)
            files.add(path)
            if "/" in item_file.path:
                dirs.add(os.path.dirname(path))

        # Delete selected files
        if not dry_run:
            self.stop()
        for path in sorted(files):
            remove_with_links(path)

        # Prune empty directories (longer paths first)
        doomed = files | dirs
        for path in sorted(dirs, reverse=True):
            residue = set(os.listdir(path) if os.path.exists(path) else [])
            ignorable = set(fnmatch.filter(residue, config.settings.CULL_WAIFS))
            if residue and residue != ignorable:
                self._engine.LOG.info(
                    "Keeping non-empty directory '%s' with %d %s%s!",
                    path,
                    len(residue),
                    "entry" if len(residue) == 1 else "entries",
                    f" ({len(ignorable)} ignorable)" if ignorable else "",
                )
            else:
                for waif in ignorable:  # - doomed:
                    waif = os.path.join(path, waif)
                    self._engine.LOG.debug(f"Deleting waif '{waif}'")
                    if not dry_run:
                        try:
                            os.remove(waif)
                        except OSError as exc:
                            self._engine.LOG.warning(
                                f"Problem deleting waif '{waif}' ({exc})"
                            )

                doomed.update(remove_with_links(path))

        # Delete item from engine
        if not dry_run:
            self.delete()

    def flush(self):
        """Write volatile data to disk."""
        self._make_it_so("saving session data of", ["d.save_resume"])


class RtorrentEngine:
    """The rTorrent backend proxy."""

    # Bare minimum fields to prefetch
    BASE_FIELDS = {
        "d.hash",
    }

    # rTorrent names of fields that never change
    CONSTANT_FIELDS = BASE_FIELDS | {
        "d.name",
        "d.is_private",
        "d.is_multi_file",
        "d.tracker_size",
        "d.size_bytes",
    }

    # rTorrent names of fields that need to be pre-fetched
    CORE_FIELDS = CONSTANT_FIELDS | {
        "d.complete",
        "d.tied_to_file",
    }

    # rTorrent names of fields that get fetched in multi-call
    PREFETCH_FIELDS = CORE_FIELDS | {
        "d.base_path",
        "d.custom=memo_alias",
        "d.custom=tm_completed",
        "d.custom=tm_loaded",
        "d.custom=tm_started",
        "d.down.rate",
        "d.down.total",
        "d.is_active",
        "d.is_open",
        "d.message",
        "d.ratio",
        "d.up.rate",
        "d.up.total",
    }

    def __init__(self, uri=None, auto_open=False):
        """Initialize proxy."""
        self.LOG = pymagic.get_class_logger(self)
        self.engine_id = "N/A"  # ID of the instance we're connecting to
        self.engine_software = "rTorrent"  # Name and version of software
        self.versions = (None, None)
        self.version_info = (0,)
        self.startup = time.time()
        self.rpc = None
        self.properties = {}
        self._item_cache = {}
        self.known_throttle_names = {"", "NULL"}
        if uri is None:
            config.autoload_scgi_url()
        else:
            config.settings.SCGI_URL = uri
        if auto_open:
            self.open()

    def view(self, viewname="default", matcher=None):
        """Get list of download items."""
        return engine.TorrentView(self, viewname, matcher)

    def group_by(self, fields, items=None):
        """Returns a dict of lists of items, grouped by the given fields.

        ``fields`` can be a string (one field) or an iterable of field names.
        """
        result = defaultdict(list)
        if items is None:
            items = self.items()

        try:
            key = operator.attrgetter(fields + "")
        except TypeError:

            def key(obj, names=tuple(fields)):
                "Helper to return group key tuple"
                return tuple(getattr(obj, x) for x in names)

        for item in items:
            result[key(item)].append(item)

        return result

    def __repr__(self):
        """Return a representation of internal state."""
        if self.rpc:
            # Connected state
            return "{} connected to {} [{}, up {}] via {!r}".format(
                self.__class__.__name__,
                self.engine_id,
                self.engine_software,
                fmt.human_duration(self.uptime, 0, 2, True).strip(),
                config.settings.SCGI_URL,
            )
        # Unconnected state
        return "{} connectable via {!r}".format(
            self.__class__.__name__,
            config.settings.SCGI_URL,
        )

    @property
    def uptime(self):
        """rTorrent's uptime."""
        return time.time() - self.startup

    def _resolve_viewname(self, viewname: str) -> str:
        """Check for special view names and return existing rTorrent one."""
        if viewname == "-":
            try:
                # Only works with rTorrent-PS at this time!
                viewname = self.open().ui.current_view()
            except rpc.ERRORS as exc:
                raise error.EngineError(f"Can't get name of current view: {exc}")

        return viewname

    def system_multicall(self, methods: Dict[str, List]) -> Dict:
        """Helper method for system.multicall"""
        results = {}
        call = []
        for method, params in methods.items():
            call.append({"methodName": method, "params": params})
        for index, r in enumerate(self.rpc.system.multicall(call)):
            if len(r) == 1:
                result = r[0]
            results[list(methods.keys())[index]] = result
        return results

    def open(self):
        """Open connection."""
        # Only connect once
        if self.rpc is not None:
            return self.rpc

        # Reading abilities are on the downfall, so...
        if not config.settings.SCGI_URL:
            raise error.UserError(
                "You need to configure a RPC connection, read"
                " https://pyrosimple.readthedocs.io/en/latest/setup.html"
            )

        # Connect and get instance ID (also ensures we're connectable)
        self.rpc = rpc.RTorrentProxy(config.settings.SCGI_URL)
        self.properties = self.system_multicall(
            {
                "system.client_version": [],
                "system.library_version": [],
                "system.time_usec": [],
                "session.name": [],
                "directory.default": [],
                "session.path": [],
            }
        )
        self.engine_id = self.properties["session.name"]
        time_usec = self.properties["system.time_usec"]

        # Make sure xmlrpc-c works as expected
        if time_usec < 2**32:
            self.LOG.warning(
                "Your xmlrpc-c is broken (64 bit integer support missing,"
                " %r returned instead)",
                type(time_usec),
            )

        # Get other manifest values
        self.engine_software = f"rTorrent {self.properties['system.library_version']}/{self.properties['system.client_version']}"

        if "+ssh:" in config.settings.SCGI_URL:
            self.startup = int(self.rpc.startup_time() or time.time())
        else:
            lockfile = os.path.join(self.properties["session.path"], "rtorrent.lock")
            try:
                self.startup = int(
                    self.rpc.execute.capture("", ["stat", "-c", "%Y", lockfile])
                )
            except (ValueError, xmlrpclib.Fault):
                self.startup = time.time()

        # Return connection
        self.LOG.debug("%s", repr(self))
        return self.rpc

    def multicall(self, viewname: str, fields: List[str]) -> Bunch:
        """Query the given fields of items in the given view.

        The result list contains named tuples,
        so you can access the fields directly by their name.
        """
        commands = tuple(f"d.{x}=" for x in fields)
        items = self.open().d.multicall2("", viewname, *commands)
        return Bunch(dict(zip([x.replace(".", "_") for x in fields], items)))

    def log(self, msg: str):
        """Log a message in the torrent client."""
        self.open().log(rpc.NOHASH, msg)

    def item(self, infohash: str, prefetch=None):
        """Fetch a single item by its info hash."""
        return next(self.items(infohash, prefetch))

    def items(
        self,
        view: Optional[Union[engine.TorrentView, str]] = None,
        prefetch: Optional[Set[str]] = None,
    ):
        """Get list of download items.

        @param view: Name of the view.
        @param prefetch: Optional list of field names to fetch initially.
        """

        multi_args: List
        if view is None:
            view = engine.TorrentView(self, "default")
        elif isinstance(view, str):
            view = engine.TorrentView(self, self._resolve_viewname(view))
        else:
            view.viewname = self._resolve_viewname(view.viewname)

        # Map pyroscope names to rTorrent ones
        if prefetch:
            prefetch = self.BASE_FIELDS | set(prefetch)
        else:
            prefetch = self.PREFETCH_FIELDS

        # Fetch items
        items = []
        try:
            # Prepare multi-call arguments
            args = sorted(prefetch)

            infohash = view._check_hash_view()
            if infohash:
                multi_call = self.open().system.multicall
                multi_args = [
                    dict(
                        methodName=field.rsplit("=", 1)[0],
                        params=[infohash]
                        + (field.rsplit("=", 1)[1].split(",") if "=" in field else []),
                    )
                    for field in args
                ]
                raw_items = [[i[0] for i in multi_call(args)]]
            else:
                multi_call = self.open().d.multicall2
                multi_args = ["", view.viewname] + [
                    field if "=" in field else field + "=" for field in args
                ]
                if view.matcher and config.settings.get("FAST_QUERY"):
                    pre_filter = matching.unquote_pre_filter(view.matcher.pre_filter())
                    self.LOG.info("!!! pre-filter: %s", pre_filter or "N/A")
                    if pre_filter:
                        multi_call = self.open().d.multicall.filtered
                        multi_args.insert(2, pre_filter)
                raw_items = multi_call(*tuple(multi_args))

            self.LOG.debug(
                "Got %d items with %d attributes from %r [%s]",
                len(raw_items),
                len(prefetch),
                self.engine_id,
                multi_call,
            )

            for item in raw_items:
                ritem = RtorrentItem(
                    self,
                    fields={},
                    rpc_fields=dict(zip(args, item)),
                )

                if view.matcher:
                    if view.matcher.match(ritem):
                        items.append(ritem)
                        yield items[-1]
                else:
                    items.append(ritem)
                    yield items[-1]
        except rpc.ERRORS as exc:
            raise error.EngineError(
                f"While getting download items from {self!r}: {exc}"
            )

    def show(
        self,
        items,
        view: Optional[str] = None,
        append: bool = False,
        disjoin: bool = False,
    ):
        """Visualize a set of items (search result), and return the view name."""
        proxy = self.open()
        view_name: str = self._resolve_viewname(view or "rtcontrol")

        if append and disjoin:
            raise error.EngineError(
                f"Cannot BOTH append to / disjoin from view '{view_name}'"
            )

        # Add view if needed
        if view not in proxy.view.list():
            proxy.view.add(rpc.NOHASH, view_name)

        # Clear view and show it
        if not append and not disjoin:
            proxy.view.filter(rpc.NOHASH, view_name, "false=")
            proxy.d.multicall2(rpc.NOHASH, "default", "d.views.remove=" + view_name)
        proxy.ui.current_view.set(rpc.NOHASH, view_name)

        # Add items
        # TODO: should be a "system.multicall"
        for item in items:
            if disjoin:
                proxy.d.views.remove(item.hash, view_name)
                proxy.view.set_not_visible(item.hash, view_name)
            else:
                proxy.d.views.push_back_unique(item.hash, view_name)
                proxy.view.set_visible(item.hash, view_name)

        return view



"""Handles RPC methods over various transports"""
import io
import logging
import socket
import subprocess
import sys
import urllib.request

from typing import Dict, List, Tuple, Type
from urllib import parse as urlparse
from urllib.error import URLError
from xmlrpc import client as xmlrpclib


logger = logging.getLogger(__name__)


def register_scheme(scheme: str):
    """Helper method to register protocols with urllib"""
    for method in filter(lambda s: s.startswith("uses_"), dir(urlparse)):
        getattr(urlparse, method).append(scheme)


class SCGIException(Exception):
    """SCGI protocol error"""


# Types of exceptions thrown
ERRORS = (SCGIException, URLError, xmlrpclib.Fault, socket.error)


class RTorrentTransport(xmlrpclib.Transport):
    """Base class for handle transports. Primarily exists to allow
    using the same transport with a different underlying RPC mechanism"""

    def __init__(self, *args, uri, codec=xmlrpclib, headers=(), **kwargs):
        self.uri = uri
        self.codec = codec
        self.verbose = False
        # We need to handle the headers differently based on the RPC protocols
        self._headers = list(headers)
        # Pass them to the transport in py 3.8+ only
        if sys.version_info.minor >= 8:
            kwargs["headers"] = self._headers
        super().__init__(*args, **kwargs)

    def parse_response(self, response):
        if self.codec == xmlrpclib:
            return super().parse_response(response)
        return self.codec.loads(response.read())


class SSHTransport(RTorrentTransport):
    """Transport via SSH conneection."""

    def request(self, host, handler, request_body, verbose=False):
        self.verbose = verbose
        target = urlparse.urlparse(self.uri).path
        cmd = ["ssh", host, "socat", "STDIO", target[1:]]
        resp = subprocess.run(
            cmd,
            input=_encode_payload(request_body, self._headers),
            capture_output=True,
            check=False,
        )
        if resp.returncode > 0:
            print("SSH command returned non-zero exit code")
            print("stderr:", resp.stderr)
            print("stdout:", resp.stdout)
        return self.parse_response(io.BytesIO(_parse_response(resp.stdout)[0]))


class HTTPTransport(RTorrentTransport):
    """Transport via HTTP(s) call."""

    def request(self, host, handler, request_body, verbose=False):
        self.verbose = verbose
        req = urllib.request.Request(self.uri)
        for key, val in self._headers:
            req.add_header(key, val)
        req.data = request_body
        with urllib.request.urlopen(req) as resp:
            return self.parse_response(resp)


class TCPTransport(RTorrentTransport):
    """Transport via TCP socket."""

    def request(self, host, handler, request_body, verbose=False):
        self.verbose = verbose
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            host, port = host.split(":")
            sock.connect((host, int(port)))
            sock.sendall(_encode_payload(request_body, self._headers))
            with sock.makefile("rb") as handle:
                return self.parse_response(
                    io.BytesIO(_parse_response(handle.read())[0])
                )


class UnixTransport(RTorrentTransport):
    """Transport via UNIX domain socket."""

    def request(self, host, handler, request_body, verbose=False):
        self.verbose = verbose
        with socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM) as sock:
            sock.connect(host)
            sock.sendall(_encode_payload(request_body))
            with sock.makefile("b") as handle:
                return self.parse_response(
                    io.BytesIO(_parse_response(handle.read())[0])
                )


TRANSPORTS = {
    "scgi": TCPTransport,
    "http": HTTPTransport,
    "https": HTTPTransport,
    "scgi+unix": UnixTransport,
    "scgi+ssh": SSHTransport,
}


for t in TRANSPORTS:
    register_scheme(t)


def transport_from_url(url: str) -> Type[xmlrpclib.Transport]:
    """Create a transport for the given URL."""
    if "/" not in url and ":" in url and url.rsplit(":")[-1].isdigit():
        url = "scgi://" + url
    elif url.startswith("/") or url.startswith("~"):
        url = "scgi+unix://"
    parsed_url = urlparse.urlsplit(url, scheme="scgi", allow_fragments=False)

    try:
        transport = TRANSPORTS[parsed_url.scheme.lower()]
    except KeyError:
        # pylint: disable=raise-missing-from
        if not any((parsed_url.netloc, parsed_url.query)) and parsed_url.path.isdigit():
            # Support simplified "domain:port" URLs
            return transport_from_url(f"scgi://{parsed_url.scheme}:{parsed_url.path}")
        raise URLError(f"Unsupported scheme in URL {parsed_url.geturl()}")
    return transport


def _encode_netstring(data: bytes) -> bytes:
    "Encode data as netstring."
    return b"%d:%s," % (len(data), data)


def _encode_headers(headers: List[Tuple[str, str]]) -> bytes:
    "Make SCGI header bytes from list of tuples."
    return b"".join(
        [b"%s\0%s\0" % (k.encode("ascii"), v.encode("ascii")) for k, v in headers]
    )


def _encode_payload(data: bytes, headers: List[Tuple[str, str]] = None) -> bytes:
    "Wrap data in an SCGI request."
    prolog: bytes = b"CONTENT_LENGTH\0%d\0SCGI\x001\0" % len(data)
    if headers:
        prolog += _encode_headers(headers)

    return _encode_netstring(prolog) + data


def _parse_headers(headers: bytes) -> Dict[str, str]:
    """
    Get headers dict from header bytestring.

    :param headers: An SCGI header bytestring
    :type headers: bytes
    :return: A dictionary of string keys/values
    """
    try:
        result = {}
        for line in headers.splitlines():
            if line:
                key, value = line.rstrip().split(b": ", 1)
                result[key.decode("ascii")] = value.decode("ascii")
        return result
    except (TypeError, ValueError) as exc:
        raise SCGIException(f"Error in SCGI headers {headers.decode()}") from exc


def _parse_response(resp: bytes) -> Tuple[bytes, Dict[str, str]]:
    """
    Get xmlrpc response from scgi response

    :param headers: An SCGI bytestring payload
    :type headers: bytes
    :return: A tuple of a binary payload and a dictionary of string keys/values
    """

    # Assume they care for standards and send us CRLF (not just LF)
    try:
        headers, payload = resp.split(b"\r\n\r\n", 1)
    except (TypeError, ValueError) as exc:
        raise SCGIException(
            f"No header delimiter in SCGI response of length {len(resp)}"
        ) from exc
    parsed_headers = _parse_headers(headers)

    clen = parsed_headers.get("Content-Length")
    if clen is not None:
        # Check length, just in case the transport is bogus
        assert len(payload) == int(clen)

    return payload, parsed_headers



""" Python Utility Functions.

    Copyright (c) 2009, 2010 The PyroScope Project <pyroscope.project@gmail.com>
"""


import json
import logging


def import_name(module_spec, name=None):
    """Import identifier C{name} from module C{module_spec}.

    If name is omitted, C{module_spec} must contain the name after the
    module path, delimited by a colon (like a setuptools entry-point).

    @param module_spec: Fully qualified module name, e.g. C{x.y.z}.
    @param name: Name to import from C{module_spec}.
    @return: Requested object.
    @rtype: object
    """
    # Hijack requests for pyrocore
    if "pyrocore" in module_spec or "pyrobase" in module_spec:
        module_spec = module_spec.replace("pyrocore", "pyrosimple")
        module_spec = module_spec.replace("pyrobase", "pyrosimple")
    # Load module
    module_name = module_spec
    if name is None:
        try:
            module_name, name = module_spec.split(":", 1)
        except ValueError:
            # pylint: disable=raise-missing-from
            raise ValueError(
                "Missing object specifier in %r (syntax: 'package.module:object.attr')"
                % (module_spec,)
            )

    try:
        module = __import__(module_name, globals(), {}, [name])
    except ImportError as exc:
        raise ImportError(f"Bad module name in {module_spec!r} ({exc})") from exc

    # Resolve the requested name
    result = module
    for attr in name.split("."):
        result = getattr(result, attr)

    return result


def get_class_logger(obj):
    """Get a logger specific for the given object's class."""
    return logging.getLogger(obj.__class__.__module__ + "." + obj.__class__.__name__)


class JSONEncoder(json.JSONEncoder):
    """Custon JSON encoder."""

    def default(self, o):
        """Support more object types."""
        if isinstance(o, set):
            return list(sorted(o))
        if hasattr(o, "as_dict"):
            return o.as_dict()
        return super().default(o)



""" Administration Tool.

    Copyright (c) 2010 The PyroScope Project <pyroscope.project@gmail.com>
"""


import pyrosimple

from pyrosimple import config
from pyrosimple.scripts.base import ScriptBase, ScriptBaseWithConfig


class AdminTool(ScriptBaseWithConfig):
    """Support for administrative tasks."""

    # TODO: config create, dump, set, get
    # TODO: backup session/config

    def add_options(self):
        super().add_options()
        subparsers = self.parser.add_subparsers()
        config_parser = subparsers.add_parser("config")
        config_parser.set_defaults(func=self.config)
        config_parser.add_argument(
            "--check", help="Check config for any issues", action="store_true"
        )

    def config(self):
        """Handle the config subcommand"""
        if self.args.check:
            try:
                config.autoload_scgi_url()
            except Exception:
                self.LOG.error("Error loading SCGI URL:")
                raise
            else:
                self.LOG.debug("Loaded SCGI URL successfully")
            try:
                pyrosimple.connect().open()
            except ConnectionRefusedError:
                self.LOG.warning(
                    "SCGI URL '%s' found, but rTorrent may not be running!",
                    config.autoload_scgi_url(),
                )
                raise
            else:
                self.LOG.debug("Connected to rTorrent successfully")

    def mainloop(self):
        self.args = self.parser.parse_args()
        self.args.func()


def run():  # pragma: no cover
    """The entry point."""
    ScriptBase.setup()
    AdminTool().run()


if __name__ == "__main__":
    run()



"""A simple job to emulate rtcontrol's actions"""
from pyrosimple import error
from pyrosimple.job.base import MatchableJob
from pyrosimple.torrent import formatting, rtorrent


class Action(MatchableJob):
    """Run an action for each matched item"""

    def __init__(self, config=None):
        super().__init__(config)
        self.action = self.config["action"]
        if not hasattr(rtorrent.RtorrentItem, self.action):
            raise error.ConfigurationError(f"Action '{self.action}' not found!")
        self.args = [formatting.env.from_string(a) for a in self.config.get("args", [])]

    def run_item(self, item):
        """For now, simply call the named methods on the item"""
        action = self.config["action"]
        processed_args = [formatting.format_item(a, item) for a in self.args]
        if self.config["dry_run"]:
            self.log.info(
                "Would %s(%s) %s",
                action,
                "" if not processed_args else processed_args,
                item.hash,
            )
        else:
            getattr(item, self.action)(*processed_args)



""" Configuration.

    For details, see https://pyrosimple.readthedocs.io/en/latest/setup.html

    Copyright (c) 2009, 2010, 2011 The PyroScope Project <pyroscope.project@gmail.com>
"""


import functools
import logging
import urllib

from pathlib import Path
from typing import Optional, Union

from dynaconf import Dynaconf, Validator

from pyrosimple import error


settings = Dynaconf(
    settings_files=[Path("~/.config/pyrosimple/config.toml").expanduser()],
    envvar="PYRO_CONF",
    envvar_prefix="PYRO",
    validators=[
        # Top-level settings
        Validator("RTORRENT_RC", default="~/.rtorrent.rc"),
        Validator("CONFIG_PY", default="~/.config/pyrosimple/config.py"),
        Validator("SORT_FIELDS", default="name,hash"),
        Validator("FAST_QUERY", gte=0, lte=2, default=0),
        Validator("SCGI_URL", default=""),
        # TOML sections
        Validator("ALIASES", default={}),
        Validator("CONNECTIONS", default={}),
        # Allow individual overrides in FORMATS section
        Validator(
            "FORMATS__default",
            default='{{d.name}} \t[{{d.alias}}]\n  {{d.is_private|fmt("is_private")}} {{d.is_open|fmt("is_open")}} {{d.is_active|fmt("is_active")}} P{{d.prio}} {%if d.is_complete %}     done{%else%}{{"%8.2f"|format(d.done)}}%{%endif%}\t{{d.size|sz}} U:{{d.up|sz}}/s  D:{{d.down|sz}}/s T:{{d.throttle|fmt("throttle")}}',
        ),
        Validator(
            "FORMATS__short",
            default='{%set ESC = "\x1B" %}{%if d.down > 0%}{{ESC+"[1m"}}{%endif%}{%if d.is_open%}O{%else%} {%endif%}{%if  d.is_active%}A{%else%} {%endif%}{%if not d.is_complete%}{{ESC+"[36m"}}{{ "{:>3}".format(d.done | round | int) }}{{ESC+"[0m"}}{%else%}  D{%endif%} {{"{:>10}".format(d.size | filesizeformat(True))}} {%if d.message%}{{ESC+"[31m"}}{%endif%} {{d.alias.rjust(3)}}{{ESC+"[0m"}} {%if d.down > 0%}{{ESC+"[1m"}}{%endif%}{{d.name}}{{ESC+"[0m"}}',
        ),
        Validator(
            "FORMATS__filelist",
            default="{% for f in d.files %}{{d.realpath}}{% if d.is_multi_file %}/{{f.path}}{% endif %}{% if loop.index != loop.length %}\n{% endif %}{% endfor %}",
        ),
        Validator(
            "FORMATS__action",
            default="{{now()|iso}} {{action}}\t {{d.name}} [{{d.alias}}]",
        ),
    ],
)


def scgi_url_from_rtorrentrc(rcfile: Union[str, Path]) -> Optional[str]:
    """Parse a rtorrent.rc file and"""
    log = logging.getLogger(__name__)
    log.debug("Loading rtorrent config from '%s'", rcfile)
    scgi_local: str = ""
    scgi_port: str = ""
    rcfile = Path(rcfile)
    with rcfile.open("r", encoding="utf-8") as handle:
        continued = False
        for line in handle.readlines():
            # Skip comments, continuations, and empty lines
            line = line.strip()
            continued, was_continued = line.endswith("\\"), continued
            if not line or was_continued or line.startswith("#"):
                continue

            # Be lenient about errors, after all it's not our own config file
            try:
                key, val = line.split("=", 1)
            except ValueError:
                log.debug("Ignored invalid line %r in %r!", line, rcfile)
                continue
            key, val = key.strip(), val.strip()

            # Copy values we're interested in
            if key in ["network.scgi.open_port", "scgi_port"]:
                log.debug("rtorrent.rc: %s = %s", key, val)
                scgi_port = val
            if key in ["network.scgi.open_local", "scgi_local"]:
                log.debug("rtorrent.rc: %s = %s", key, val)
                scgi_local = val

    # Validate fields
    if scgi_local and not scgi_port.startswith("scgi+unix://"):
        scgi_local = "scgi+unix://" + str(Path(scgi_local).expanduser())
    if scgi_port and not scgi_port.startswith("scgi://"):
        scgi_port = "scgi://" + scgi_port

    return scgi_local or scgi_port
    # Prefer UNIX domain sockets over TCP sockets


def autoload_scgi_url() -> str:
    """Load and return the SCGI URL, auto-resolving it if necessary."""
    if settings.SCGI_URL:
        return str(settings.SCGI_URL)
    # Get and check config file name
    rcfile = Path(settings.RTORRENT_RC).expanduser()
    if not rcfile.exists():
        raise error.UserError(f"rTorrent RC file '{rcfile}' doesn't exist!")
    scgi_url = scgi_url_from_rtorrentrc(rcfile)

    settings.set("SCGI_URL", scgi_url)

    return str(settings.SCGI_URL)


def lookup_announce_alias(name: str):
    """Get canonical alias name and announce URL list for the given alias."""
    for alias, urls in settings["ALIASES"].items():
        if alias.lower() == name.lower():
            return alias, urls

    raise KeyError(f"Unknown alias {name}")


def lookup_announce_url(name: str):
    """Get canonical alias name and announce URL list for the given alias.

    Unlike lookup_announce_alias, only valid URLs are returned"""
    for alias, urls in settings["ALIASES"].items():
        if alias.lower() == name.lower():
            result = []
            for url in urls:
                if urllib.parse.urlparse(url).scheme:
                    result.append(url)
            return alias, result

    raise KeyError(f"Unknown alias {name}")


@functools.lru_cache(maxsize=None)
def map_announce2alias(url: str) -> str:
    """Get tracker alias for announce URL, and if none is defined, the 2nd level domain."""
    if url in settings["ALIASES"].items():
        return url
    # Try to find an exact alias URL match and return its label
    for alias, urls in settings["ALIASES"].items():
        if any(i == url for i in urls):
            return str(alias)

    # Try to find an alias URL prefix and return its label
    parts = urllib.parse.urlparse(url)
    server = urllib.parse.urlunparse(
        (parts.scheme, parts.netloc, "/", None, None, None)
    )

    for alias, urls in settings["ALIASES"].items():
        if any(i.startswith(server) for i in urls):
            return str(alias)

    # Return 2nd level domain name if no alias found
    try:
        # Try to find based on domain
        domain = ".".join(parts.netloc.split(":")[0].split(".")[-2:])
        for alias, urls in settings["ALIASES"].items():
            if any(i == domain for i in urls):
                return str(alias)
        return domain
    except IndexError:
        return parts.netloc


py_loaded = False


def load_custom_py():
    """Load custom python configuration.

    This only gets called when CLI tools are called to prevent some weird code injection"""
    if py_loaded:
        return
    log = logging.getLogger(__name__)
    config_file = Path(settings.CONFIG_PY).expanduser()
    if config_file.exists():
        log.debug("Loading '%s'...", config_file)
        with open(config_file, "rb") as handle:
            # pylint: disable=exec-used
            exec(handle.read())
    else:
        log.debug("Configuration file '%s' not found!", config_file)



""" Color theme support.

    Copyright (c) 2017 The PyroScope Project <pyroscope.project@gmail.com>
"""


import glob
import os
import sys

from pyrosimple.scripts.base import ScriptBase, ScriptBaseWithConfig


class ThemeSwitcher(ScriptBaseWithConfig):
    """Rotate through color themes."""

    # argument description for the usage information
    ARGS_HELP = ""

    def add_options(self):
        """Add program options."""
        super().add_options()

        self.add_bool_option("-l", "--list", help="list available themes")
        self.add_bool_option(
            "-c", "--current", help="print path to currently selected theme"
        )
        self.add_bool_option(
            "-n", "--next", help="rotate through selected themes, and print new path"
        )
        self.add_bool_option(
            "-a", "--all", help="remove any selections, and use all themes"
        )
        self.add_value_option(
            "-t", "--toggle", "NAME", help="toggle selection of a theme"
        )

    def mainloop(self):
        """Handle theme selection changes, or rotate through selection."""
        themes_dir = os.path.join(self.config_dir, "color-schemes")
        selected_file = os.path.join(themes_dir, ".selected")
        current_file = os.path.join(themes_dir, ".current")

        # Read persisted state
        selected_themes = []
        if os.path.exists(selected_file):
            with open(selected_file, encoding="ascii") as handle:
                selected_themes = [x.strip() for x in handle]

        current_theme = None
        if os.path.exists(current_file):
            with open(current_file, encoding="ascii") as handle:
                current_theme = handle.readline().strip()

        # Scan config for available themes
        themes = {}
        for ext in (".rc.default", ".rc"):
            for filepath in glob.glob(themes_dir + "/*" + ext):
                name = os.path.basename(filepath).split(".")[0]
                if name:
                    # This prefers '.rc.default' files, because those are checked first
                    themes.setdefault(name, filepath)

        # Use available selected themes in given order, if there are any, else all themes
        if selected_themes and not set(selected_themes).isdisjoint(set(themes)):
            theme_list = [x for x in selected_themes if x in themes]
        else:
            theme_list = list(sorted(themes))

        # Check options
        if self.options.list or self.options.all or self.options.toggle:
            if self.options.all:
                if os.path.exists(selected_file):
                    os.remove(selected_file)
                selected_themes = []

            for name in (self.options.toggle or "").replace(",", " ").split():
                if name not in themes:
                    self.parser.error(
                        f"Unknown theme {name!r}, use '--list' to show them"
                    )
                elif name in selected_themes:
                    selected_themes = [x for x in selected_themes if x != name]
                else:
                    selected_themes.append(name)

                with open(selected_file, "wt", encoding="ascii") as handle:
                    handle.write("\n".join(selected_themes + [""]))

            if self.options.list:
                for name in sorted(themes):
                    print(
                        "{} {} {}".format(
                            "*" if name == current_theme else " ",
                            f"{selected_themes.index(name) + 1:2d}"
                            if name in selected_themes
                            else "  ",
                            name,
                        )
                    )

        elif self.options.current or self.options.next:
            # Determine current theme, or rotate to next one
            new_theme = theme_list[0]
            if self.options.current and current_theme in theme_list:
                new_theme = current_theme
            elif current_theme in theme_list:
                new_theme = (theme_list * 2)[theme_list.index(current_theme) + 1]

            # Persist new theme
            if new_theme != current_theme:
                with open(current_file, "wt", encoding="ascii") as handle:
                    handle.write(new_theme + "\n")

            # Return result
            sys.stdout.write(themes[new_theme])
            sys.stdout.flush()

        else:
            self.LOG.info(
                "Current color theme is '{}'.".format(
                    current_theme if current_theme in theme_list else theme_list[0]
                )
            )
            self.LOG.info("Use '--help' to get usage information.")


def run():  # pragma: no cover
    """The entry point."""
    ScriptBase.setup()
    ThemeSwitcher().run()


if __name__ == "__main__":
    run()



"""Helper classes to deal with caching various data"""
import time

from collections import abc
from threading import RLock
from typing import Dict, Optional


class ExpiringCache(abc.MutableMapping):
    """Caches items for a fixed time, with an optional exlusionary list"""

    def __init__(self, items: Optional[Dict] = None, expires=5, static_keys=None):
        self.expires = expires
        # When 3.7 support is dropped, the type can be made more specific:
        # Dict[str, tuple[float, Any]]
        self.data: Dict[str, tuple] = {}
        self.lock = RLock()
        self.static_keys = static_keys or set()
        if items:
            for key, val in items:
                if key in self.static_keys:
                    expire_at = 0
                else:
                    expire_at = self.expires + time.time()
                self.data[key] = (expire_at, val)

    def __delitem__(self, key):
        with self.lock:
            del self.data[key]

    def __setitem__(self, key, val, expire=None):
        with self.lock:
            if key in self.static_keys or expire == 0:
                expire_at = 0
            else:
                expire_at = (expire or self.expires) + time.time()
            self.data[key] = (expire_at, val)

    def __getitem__(self, key, with_age=False):
        with self.lock:
            expires_at, item = self.data[key]
            if expires_at == 0 or expires_at > time.time():
                if with_age:
                    return item, expires_at - time.time()
                return item
            del self[key]
            raise KeyError(key)

    def __contains__(self, key):
        """Return True if the dict has a key, else return False."""
        try:
            with self.lock:
                expires_at, _ = self.data.get(key)
                if expires_at == 0 or expires_at < time.time():
                    return True
                del self[key]
        except KeyError:
            pass
        return False

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        return self.data.__iter__()



"""Move torrents to a new path.
Primarily intended for move-on-completion, but is intended to be generic.
"""
import concurrent.futures
import shutil
import time

from pathlib import Path

import pyrosimple

from pyrosimple import error
from pyrosimple.torrent import formatting, rtorrent
from pyrosimple.util import matching, pymagic, rpc


def move(i: rtorrent.RtorrentItem, target: str):
    """Move a torrent path"""
    i.stop()
    for _ in range(0, 5):
        if i.rpc_call("d.is_open", cache=False):
            time.sleep(0.1)
        else:
            break
    shutil.move(i.datapath(), target)
    i._engine.rpc.d.directory.set(i.hash, target)
    i.start()


class PathMover:
    """Conditionally move torrent paths.
    Currently it just relies on the current datapath"""

    def __init__(self, config=None):
        """Set up statistics logger."""
        self.config = config or {}
        self.config.setdefault("max_workers", 1)
        self.config.setdefault("dry_run", False)
        if not self.config["target"]:
            raise Exception("'target' not defined!")
        self.engine = None
        self.LOG = pymagic.get_class_logger(self)
        self.LOG.debug("Path mover created with config %r", self.config)

    def check_and_move(self, i: rtorrent.RtorrentItem):
        """Conditionally move data"""
        target = formatting.format_item_str(self.config["target"], i)
        if not target:
            self.LOG.debug("Empty target for %s", i.hash)
            return
        if (Path(i.fetch("directory")) == Path(target)) or (
            i.rpc_call("d.is_multi_file")
            and Path(i.fetch("directory")).parent == Path(target)
        ):
            self.LOG.debug("%s already moved, skipping", i.hash)
            return
        if self.config["dry_run"]:
            self.LOG.info(
                "Would move %s from '%s' to '%s'", i.hash, i.datapath(), target
            )
            return
        self.LOG.info(
            "Moving path for %s from '%s' to '%s'", i.hash, i.datapath(), target
        )
        move(i, target)
        i.flush()

    def run(self):
        """Check if any torrents need to be moved"""
        try:
            self.engine = pyrosimple.connect()
            matcher = matching.create_matcher(self.config["matcher"])
            futures = []
            with concurrent.futures.ThreadPoolExecutor(
                max_workers=self.config["max_workers"]
            ) as executor:
                # Submit tasks
                for i in self.engine.view("default", matcher):
                    if matcher.match(i):
                        futures.append(executor.submit(self.check_and_move, i))
                # Wait and check for exceptions
                for future in concurrent.futures.as_completed(futures):
                    exc = future.result()
                    if exc is not None:
                        self.LOG.error("Could not move: %s: %s", future, exc)
        except (error.LoggableError, *rpc.ERRORS) as exc:
            self.LOG.warning(str(exc))



def _custom_fields():
    """Yield custom field definitions."""
    # Import some commonly needed modules
    import os

    from pyrosimple.torrent import engine, matching
    from pyrosimple.util import fmt

    # PUT CUSTOM FIELD CODE HERE
    # Disk space check (as an example)
    # see https://pyrosimple.readthedocs.io/en/latest/custom.html#has-room
    def has_room(obj):
        "Check disk space."
        pathname = obj.path
        if pathname and not os.path.exists(pathname):
            pathname = os.path.dirname(pathname)
        if pathname and os.path.exists(pathname):
            stats = os.statvfs(pathname)
            return stats.f_bavail * stats.f_frsize - int(
                diskspace_threshold_mb
            ) * 1024**2 > obj.size * (1.0 - obj.done / 100.0)
        else:
            return None

    yield engine.DynamicField(
        engine.untyped,
        "has_room",
        "check whether the download will fit on its target device",
        matcher=matching.BoolFilter,
        accessor=has_room,
        formatter=lambda val: "OK" if val else "??" if val is None else "NO",
    )
    globals().setdefault("diskspace_threshold_mb", "500")


custom_field_factories.append(_custom_fields)



# pylint: disable=no-self-use,too-many-nested-blocks
""" rTorrent Control.

    Copyright (c) 2010, 2011 The PyroScope Project <pyroscope.project@gmail.com>
"""

import argparse
import functools
import json
import re
import shlex
import subprocess
import sys
import time

from multiprocessing.pool import ThreadPool
from typing import Callable, List, Union

from pyrosimple import config
from pyrosimple.scripts.base import PromptDecorator, ScriptBase, ScriptBaseWithConfig
from pyrosimple.torrent import engine, formatting, rtorrent
from pyrosimple.util import matching, pymagic, rpc
from pyrosimple.util.parts import DefaultBunch


def print_help_fields():
    """Print help about fields and field formatters."""
    # Mock entries, so they fulfill the expectations towards a field definition
    def custom_manifold():
        "named rTorrent custom attribute, e.g. 'custom_completion_target'"
        return ("custom_KEY", custom_manifold)

    def kind_manifold():
        "file types that contribute at least N% to the item's total size"
        return ("kind_N", kind_manifold)

    print("")
    print("Fields are:")
    print(
        "\n".join(
            [
                "  %-21s %s" % (name, field.__doc__)
                for name, field in sorted(
                    list(engine.FieldDefinition.FIELDS.items())
                    + [
                        custom_manifold(),
                        kind_manifold(),
                    ]
                )
            ]
        )
    )


class FieldStatistics:
    """Collect statistical values for the fields of a search result."""

    def __init__(self):
        "Initialize accumulator"
        self.size = 0
        self.errors = DefaultBunch(int)
        self.total = DefaultBunch(int)
        self.min = DefaultBunch(int)
        self.max = DefaultBunch(int)
        self._basetime = time.time()
        self.intermixed_args = True

    def __bool__(self):
        "Truth"
        return bool(self.total)

    def __nonzero__(self):
        return self.__bool__()

    def add(self, field, val):
        "Add a sample"
        if engine.FieldDefinition.FIELDS[field]._matcher is matching.TimeFilter:
            val = self._basetime - val

        try:
            self.total[field] += val
            self.min[field] = min(self.min[field], val) if field in self.min else val
            self.max[field] = max(self.max[field], val)
            self.size += 1
        except (ValueError, TypeError):
            self.errors[field] += 1

    @property
    def average(self):
        "Calculate average"
        result = DefaultBunch(str)

        # Calculate average if possible
        if self.size:
            result.update(
                (key, "" if isinstance(val, str) else val / self.size)
                for key, val in list(self.total.items())
            )

        # Handle time fields
        # for key, fielddef in  engine.FieldDefinition.FIELDS.items():
        #    if key in result and fielddef._matcher is matching.TimeFilter:
        #       result[key] = ''
        # for key, fielddef in  engine.FieldDefinition.FIELDS.items():
        #    if key in result and fielddef._matcher is matching.TimeFilter:
        #        result[key] = engine._fmt_duration(result[key])
        # print self.total
        # print result
        return result


class RtorrentAction(argparse.Action):
    """This class is used by the argparse action parameter for adding rtcontrol actions to a master list in the namespace.

    There is a rather unforunate name collision between argparse's actions and rtcontrol's actions.
    'const' is used as the method name to call, with the arguments being pulled from the value"""

    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        """Build the action, default to 1 narg since that's the most common"""
        if nargs is None:
            nargs = 1
        if "const" not in kwargs:
            kwargs["const"] = option_strings[0].lstrip("-").replace("-", "_")
        self.interactive = False
        super().__init__(option_strings, dest, nargs, **kwargs)

    def __call__(
        self, parser, namespace, values, option_string=None, interactive=False
    ):
        """Add any action to the namesapce in order"""
        actions = getattr(namespace, "actions", [])
        actions.append(
            {"method": self.const, "args": values, "interactive": interactive}
        )
        namespace.actions = actions


class RtorrentInteractiveAction(RtorrentAction):
    """Simple class to mark commands as interactive"""

    def __call__(self, *args, **kwargs):
        """Call the parent, but with interactive=True"""
        super().__call__(*args, **kwargs, interactive=True)


class RtorrentControl(ScriptBaseWithConfig):
    ### Keep things wrapped to fit under this comment... ##############################
    """
    Control and inspect rTorrent from the command line.

    Filter expressions take the form "<field>=<value>", and all expressions must
    be met (AND). If a field name is omitted, "name" is assumed. You can also use
    uppercase OR to build a list of alternative conditions.

    For numeric fields, a leading "+" means greater than, a leading "-" means less
    than. For string fields, the value is a glob pattern (*, ?, [a-z], [!a-z]), or
    a regex match enclosed by slashes. All string comparisons are case-ignoring.
    Multiple values separated by a comma indicate several possible choices (OR).
    "!" in front of a filter value negates it (NOT).

    See https://pyrosimple.readthedocs.io/en/latest/usage.html#rtcontrol for more.

    Examples:
      - All 1:1 seeds         ratio=+1
      - All active torrents   xfer=+0
      - All seeding torrents  up=+0
      - Slow torrents         down=+0 down=-5k
      - Older than 2 weeks    completed=+2w
      - Big stuff             size=+4g
      - 1:1 seeds not on NAS  ratio=+1 'realpath=!/mnt/*'
      - Music                 kind=flac,mp3
    """

    # argument description for the usage information
    ARGS_HELP = "<filter>..."

    # additonal stuff appended after the command handler's docstring
    ADDITIONAL_HELP = [
        "",
        "",
        "Use --help to get a list of all options.",
        "Use --help-fields to list all fields and their description.",
    ]

    # additional values for output formatting
    FORMATTER_DEFAULTS = dict(
        now=time.time,
    )

    # choices for --ignore
    IGNORE_OPTIONS = ("0", "1")

    # choices for --prio
    PRIO_OPTIONS = ("0", "1", "2", "3")

    # choices for --alter
    ALTER_MODES = ("append", "remove")

    # action options that perform some change on selected items
    ACTION_MODES = (
        # TODO: Bunch(name="announce", options=("--announce",), help="announce right now", interactive=True),
        # TODO: --pause, --resume?
        # TODO: implement --clean-partial
        # self.add_bool_option("--clean-partial",
        #    help="remove partially downloaded 'off'ed files (also stops downloads)")
        # TODO: --move / --link output_format / the formatted result is the target path
        #           if the target contains a '//' in place of a '/', directories
        #           after that are auto-created
        #           "--move tracker_dated", with a custom output format
        #           like "tracker_dated = ~/done//$(alias)s/$(completed).7s",
        #           will move to ~/done/OBT/2010-08 for example
        #        self.add_value_option("--move", "TARGET",
        #            help="move data to given target directory (implies -i, can be combined with --delete)")
        # TODO: --copy, and --move/--link across devices
    )

    def __init__(self):
        """Initialize rtcontrol."""
        super().__init__()

        self.prompt = PromptDecorator(self)
        self.is_plain_output_format = False
        self.original_output_format = None

    def add_options(self):
        """Add program options."""
        super().add_options()

        # basic options
        self.add_bool_option(
            "--help-fields", help="show available fields and their description"
        )
        self.add_bool_option(
            "-n", "--dry-run", help="don't commit changes, just tell what would happen"
        )
        self.prompt.add_options()

        # output control
        output_group = self.parser.add_argument_group("output")
        output_group.add_argument(
            "-S",
            "--shell",
            help="escape output following shell rules",
            action="store_true",
        )
        output_group.add_argument(
            "-0",
            "--nul",
            "--print0",
            action="store_true",
            help="use a NUL character instead of a linebreak after items",
        )
        output_group.add_argument(
            "-+",
            "--stats",
            help="add sum / avg / median of numerical fields",
            action="store_true",
        )
        output_group.add_argument(
            "--summary",
            help="print only statistical summary, without the items",
            action="store_true",
        )
        output_group.add_argument(
            "--json",
            help="dump default fields of all items as JSON (use '-o f1,f2,...' to specify fields)",
            action="store_true",
        )
        output_group.add_argument(
            "-o",
            "--output-format",
            metavar="FORMAT",
            help="specify display format (use '-o-' to disable item display)",
        )
        output_group.add_argument(
            "-O",
            "--output-template",
            metavar="FILE",
            help="pass control of output formatting to the specified template",
        )
        output_group.add_argument(
            "-s",
            "--sort-fields",
            metavar="FIELD",
            help="fields used for sorting, descending if prefixed with a '-'; '-s*' uses output field list",
        )
        output_group.add_argument(
            "-r",
            "--reverse-sort",
            help="reverse the sort order",
            action="store_true",
        )
        self.add_value_option(
            "-/",
            "--select",
            "[N-]M",
            help="select result subset by item position (counting from 1)",
        )
        self.add_bool_option(
            "-V", "--view-only", help="show search result only in default ncurses view"
        )
        self.add_value_option(
            "--to-view",
            "--to",
            "NAME",
            help="show search result only in named ncurses view",
        )
        self.add_value_option(
            "--alter-view",
            "--alter",
            "MODE",
            type="choice",
            default=None,
            choices=self.ALTER_MODES,
            help="alter view according to mode: {} (modifies -V and --to behaviour)".format(
                ", ".join(self.ALTER_MODES)
            ),
        )
        self.add_bool_option(
            "--tee-view",
            "--tee",
            help="ADDITIONALLY show search results in ncurses view (modifies -V and --to behaviour)",
        )
        self.add_value_option(
            "--from-view",
            "--from",
            "NAME",
            help="select only items that are on view NAME (NAME can be an info hash to quickly select a single item)",
        )
        self.add_value_option(
            "-M",
            "--modify-view",
            "NAME",
            help="get items from given view and write result back to it (short-cut to combine --from-view and --to-view)",
        )
        self.parser.add_argument(
            "-Q",
            "--fast-query",
            choices=("=", "0", "1", "2"),
            default="=",
            metavar="LEVEL",
            help="enable query optimization (=: use config; 0: off; 1: safe; 2: danger seeker)",
        )
        action_group = self.parser.add_argument_group(
            "actions", "Can be set more than once, and order matters."
        )
        action_group.add_argument(
            "--call",
            action=RtorrentInteractiveAction,
            help="call an OS command pattern in the shell (implies -i)",
            metavar="CMD [--call]",
        )
        action_group.add_argument(
            "--spawn",
            action=RtorrentInteractiveAction,
            help="execute OS command pattern(s) directly (implies -i)",
            metavar="CMD [--spawn ...]",
        )
        action_group.add_argument(
            "--flush",
            "-F",
            nargs=0,
            action=RtorrentAction,
            help="flush changes immediately (save session data)",
        )
        action_group.add_argument(
            "--ignore",
            action=RtorrentAction,
            choices=self.IGNORE_OPTIONS,
            help="set 'ignore commands' status on torrent",
        )
        action_group.add_argument(
            "--start",
            action=RtorrentAction,
            nargs=0,
            help="start torrent",
        )
        action_group.add_argument(
            "--stop",
            "--close",
            action=RtorrentAction,
            nargs=0,
            help="stop torrent",
        )
        action_group.add_argument(
            "--hash-check",
            "-H",
            action=RtorrentInteractiveAction,
            nargs=0,
            help="trigger a hash check (implies -i)",
        )
        action_group.add_argument(
            "--prio",
            action=RtorrentAction,
            choices=self.PRIO_OPTIONS,
            help="set priority of torrent",
        )
        action_group.add_argument(
            "--delete",
            action=RtorrentInteractiveAction,
            nargs=0,
            help="remove torrent (but not the data) (implies -i)",
        )
        action_group.add_argument(
            "--cull",
            action=RtorrentInteractiveAction,
            nargs=0,
            help="remove torrent and ALL data files (implies -i)",
        )
        action_group.add_argument(
            "--purge",
            action=RtorrentInteractiveAction,
            nargs=0,
            help="remove torrent and partial data files (implies -i)",
        )
        action_group.add_argument(
            "--throttle",
            "-T",
            action=RtorrentInteractiveAction,
            help="assign to named throttle group (NULL=unlimited, NONE=global) (implies -i)",
        )
        action_group.add_argument(
            "--tag",
            action=RtorrentInteractiveAction,
            metavar='"TAG +TAG -TAG..."',
            help="add or remove tag",
        )
        action_group.add_argument(
            "--custom",
            action=RtorrentAction,
            metavar="KEY=VALUE",
            help="set value of 'custom_KEY' field (KEY might also be 1..5)",
        )
        action_group.add_argument(
            "--exec",
            "--rpc",
            action=RtorrentInteractiveAction,
            const="execute",
            metavar="RPC_CMD",
            help="execute RPC command pattern",
        )

    def help_completion_fields(self):
        """Return valid field names."""
        for name, field in sorted(engine.FieldDefinition.FIELDS.items()):
            if issubclass(field._matcher, matching.BoolFilter):
                yield f"{name}=no"
                yield f"{name}=yes"
                continue
            if issubclass(field._matcher, matching.PatternFilter):
                yield f"{name}="
                yield f"{name}=/"
                yield f"{name}=?"
                yield f"{name}=\"'*'\""
                continue
            if issubclass(field._matcher, matching.NumericFilterBase):
                for i in range(10):
                    yield "%s=%d" % (name, i)
            else:
                yield f"{name}="

            yield rf"{name}=+"
            yield rf"{name}=-"

        yield "custom_"
        yield "kind_"

    # TODO: refactor to engine.TorrentProxy as format() method
    def format_item(self, item: str, defaults=None, stencil=None) -> str:
        """Format an item."""

        try:
            item_text: str = formatting.format_item(
                self.options.output_format_template, item, defaults
            )
        except (NameError, ValueError, TypeError) as exc:
            self.fatal(
                "Trouble with formatting item %r\n\n  FORMAT = %r\n\n  REASON ="
                % (item, self.options.output_format),
                exc,
            )
            raise  # in --debug mode

        if self.options.shell:
            item_text = "\t".join(shlex.quote(i) for i in item_text.split("\t"))

        # Justify headers according to stencil
        if stencil:
            item_text = "\t".join(
                i.ljust(len(s)) for i, s in zip(item_text.split("\t"), stencil)
            )

        return item_text

    def emit(
        self,
        item,
        defaults=None,
        stencil=None,
        to_log: Union[bool, Callable] = False,
        item_formatter=None,
    ):
        """Print an item to stdout, or the log on INFO level."""
        item_text: str = self.format_item(item, defaults, stencil)

        # Post-process line?
        if item_formatter:
            item_text = item_formatter(item_text)

        # Dump to selected target
        if to_log:
            if callable(to_log):
                to_log(item_text)
            else:
                self.LOG.info(item_text)
        elif self.options.nul:
            print(item_text, end="\0")
        else:
            print(item_text)

    # TODO: refactor to formatting.OutputMapping as a class method
    def validate_output_format(self, default_format):
        """Prepare output format for later use."""
        output_format = self.options.output_format
        self.original_output_format = output_format

        # Use default format if none is given
        if output_format is None:
            output_format = default_format

        # Check if it's a custom output format from configuration
        # (they take precedence over field names, so name them wisely)
        if output_format in config.settings.FORMATS:
            output_format = config.settings.FORMATS.get(output_format)

        # Expand plain field list to usable form
        # "name,size.sz" would become "{{d.name}}\t{{d.size|sz}}"
        if re.match(r"^[,._0-9a-zA-Z]+$", output_format):
            self.is_plain_output_format = True
            outputs = []
            for field in formatting.validate_field_list(
                output_format, allow_fmt_specs=True
            ):
                field = field.replace(".", "|")
                if len(field.split("|")) == 1:
                    outputs += [f"{{{{d.{field}|fmt('{field}')}}}}"]
                else:
                    outputs += ["{{d.%s}}" % field]
            output_format = "\t".join(outputs)

        # Replace some escape sequences
        output_format = (
            output_format.replace(r"\n", "\n")
            .replace(r"\t", "\t")
            .replace(r"\ ", " ")  # to prevent stripping in config file
        )
        self.options.output_format = output_format
        self.options.output_format_template = formatting.env.from_string(output_format)

    # TODO: refactor to engine.FieldDefinition as a class method
    def get_output_fields(self) -> List[str]:
        """Get field names from output template."""
        result = []
        for name in formatting.get_fields_from_template(self.options.output_format):
            if name not in engine.FieldDefinition.FIELDS:
                self.LOG.warning(
                    "Omitted unknown name '%s' from statistics and output format sorting",
                    name,
                )
            else:
                result.append(name)

        return result

    def validate_sort_fields(self):
        """Take care of sorting."""
        if self.options.sort_fields is None:
            self.options.sort_fields = config.settings.SORT_FIELDS
        if self.options.sort_fields == "*":
            self.options.sort_fields = self.get_output_fields()

        return formatting.validate_sort_fields(self.options.sort_fields)

    def show_in_view(self, sourceview, matches, targetname=None):
        """Show search result in ncurses view."""
        append = self.options.alter_view == "append"
        remove = self.options.alter_view == "remove"
        action_name = (
            ", appending to" if append else ", removing from" if remove else " into"
        )
        targetname = self.engine.show(
            matches,
            targetname or self.options.to_view or "rtcontrol",
            append=append,
            disjoin=remove,
        )
        msg = "Filtered %d out of %d torrents using [ %s ]" % (
            len(matches),
            sourceview.size(),
            sourceview.matcher,
        )
        self.LOG.info("%s%s rTorrent view %r.", msg, action_name, targetname)
        self.engine.log(msg)

    def mainloop(self):
        """The main loop."""
        # Print field definitions?
        if self.options.help_fields:
            self.parser.print_help()
            print_help_fields()
            sys.exit(1)

        # Print usage if no conditions are provided
        if not self.args:
            self.parser.error("No filter conditions given!")

        # Check special action options
        actions = getattr(self.options, "actions", [])

        # Reduce results according to index range
        selection = None
        if self.options.select:
            try:
                if "-" in self.options.select:
                    selection = tuple(
                        int(i or default, 10)
                        for i, default in zip(
                            self.options.select.split("-", 1), ("1", "-1")
                        )
                    )
                else:
                    selection = 1, int(self.options.select, 10)
            except (ValueError, TypeError) as exc:
                self.fatal(f"Bad selection '{self.options.select}' ({exc})")

        # Preparation steps
        if self.options.fast_query != "=":
            config.settings.set("FAST_QUERY", int(self.options.fast_query))
        raw_output_format = self.options.output_format
        default_output_format = "default"
        if actions:
            default_output_format = "action"
        self.validate_output_format(default_output_format)
        sort_key = self.validate_sort_fields()
        # This bit of magic here is to make the CLI more intuitive.
        # Without it, `rtcontrol "*Ubuntu Server*"` would become two queries:
        # `*Ubuntu` and `Server*`
        args = []
        for a in self.args:
            if "=" not in a and not a.startswith('"'):
                a = f'"{a}"'
            args.append(a)
        query_tree = matching.QueryGrammar.parse(" ".join(args))
        # Use validate_sort_fields to pre-validate key name
        key_names = matching.KeyNameVisitor().visit(query_tree)
        # Handles the empty regex query "//"
        if not key_names:
            key_names = ["name"]
        formatting.validate_sort_fields(",".join(key_names))
        matcher = matching.MatcherBuilder().visit(query_tree)
        self.LOG.debug("Matcher is: %s", matcher)

        # View handling
        if self.options.modify_view:
            if self.options.from_view or self.options.to_view:
                self.fatal(
                    "You cannot combine --modify-view with --from-view or --to-view"
                )
            self.options.from_view = self.options.to_view = self.options.modify_view

        # Holds summary information, will be populated later
        summary = FieldStatistics()

        # Find matching torrents
        engines = {}
        for uri in self.multi_connection_lookup(config.settings["SCGI_URL"]):
            engines[uri] = rtorrent.RtorrentEngine(uri, auto_open=True)
        # Kick off the result fetcher in a thread pool
        pool = ThreadPool(processes=len(engines))
        futures = {}
        for uri, r_engine in engines.items():

            def fetch(e):
                view = e.view(self.options.from_view, matcher)
                prefetch = [
                    engine.FieldDefinition.FIELDS[f].requires
                    for f in self.get_output_fields()
                    + key_names
                    + [
                        s[1:] if s.startswith("-") else s
                        for s in self.options.sort_fields.split(",")
                    ]
                ]
                prefetch = [item for sublist in prefetch for item in sublist]
                matches = list(e.items(view=view, prefetch=prefetch))
                matches.sort(key=sort_key, reverse=self.options.reverse_sort)
                return matches

            futures[uri] = pool.apply_async(fetch, (r_engine,))

        # The rest of the process should still be done in sequence
        for uri, r_engine in engines.items():
            view = r_engine.view(self.options.from_view, matcher)
            matches = futures[uri].get()

            if selection:
                matches = matches[selection[0] - 1 : selection[1]]

            if not matches:
                # Think "404 NOT FOUND", but then exit codes should be < 256
                self.return_code = 44

            # Tee to ncurses view, if requested
            if self.options.tee_view and (
                self.options.to_view or self.options.view_only
            ):
                self.show_in_view(view, matches)

            # Generate summary?
            if self.options.stats or self.options.summary:
                for field in self.get_output_fields():
                    try:
                        0 + getattr(matches[0], field)
                    except (TypeError, ValueError, IndexError):
                        summary.total[field] = ""
                    else:
                        for item in matches:
                            summary.add(field, getattr(item, field))

            # Run actions?
            if actions:
                self.LOG.info(
                    "%s perform actions [%s] on %d out of %d torrents.",
                    "Would" if self.options.dry_run else "About to",
                    ",".join([a["method"] for a in actions]),
                    len(matches),
                    view.size(),
                )
            for item in matches:
                for action in actions:
                    action_name = action["method"].replace("_", " ")
                    defaults = {"action": action_name, "now": time.time}
                    defaults.update(self.FORMATTER_DEFAULTS)

                    args = action["args"]
                    # Templatetize arguments for some commands
                    if action_name in ["call", "spawn", "execute"]:
                        template_args = [
                            ("{##}" + i if "{{" in i else i) for i in action["args"]
                        ]
                        if not self.prompt.ask_bool(f"{action_name} item {item.name}"):
                            continue
                        if (
                            self.options.output_format
                            and not self.options.view_only
                            and str(self.options.output_format) != "-"
                        ):
                            self.emit(item, defaults)

                        args = tuple(
                            formatting.format_item(
                                formatting.env.from_string(i),
                                item,
                                defaults=dict(item=item),
                            )
                            for i in template_args
                        )

                    if self.options.dry_run:
                        self.LOG.debug("Would call action %s%r", action["method"], args)
                    else:
                        if action_name == "call":
                            self.LOG.debug("Calling '%s' with a shell", args[0])
                            subprocess.run(args[0], check=True, shell=True)
                            continue
                        if action_name == "spawn":
                            args = shlex.split(args[0])
                            self.LOG.debug("Runing '%s'", args)
                            subprocess.run(args, check=True, shell=False)
                            continue
                        getattr(item, action["method"])(*args)
                        if self.options.view_only:
                            show_in_client = functools.partial(
                                lambda x, e: e.open().log(rpc.NOHASH, x), e=r_engine
                            )
                            self.emit(item, defaults, to_log=show_in_client)

            # Show in ncurses UI?
            if not self.options.tee_view and (
                self.options.to_view or self.options.view_only
            ):
                self.show_in_view(view, matches)

            # Dump as JSON array?
            elif self.options.json:
                json_data = matches
                if raw_output_format:
                    json_fields = raw_output_format.split(",")
                    json_data = [
                        {name: getattr(i, name) for name in json_fields}
                        for i in matches
                    ]
                json.dump(
                    json_data,
                    sys.stdout,
                    indent=2,
                    separators=(",", ": "),
                    sort_keys=True,
                    cls=pymagic.JSONEncoder,
                )
                sys.stdout.write("\n")
                sys.stdout.flush()

            # Show via template?
            elif self.options.output_template:
                full_ns = dict(
                    version=None,
                    proxy=r_engine.open(),
                    view=view,
                    query=matcher,
                    matches=matches,
                    summary=summary,
                )

                output_template = self.options.output_template
                sys.stdout.write(formatting.expand_template(output_template, full_ns))
                sys.stdout.flush()

            # Show on console?
            elif self.options.output_format and str(self.options.output_format) != "-":
                if not self.options.summary:
                    for item in matches:
                        # Print matching item
                        self.emit(item, self.FORMATTER_DEFAULTS)

                # Print summary?
                if matches and summary:
                    print(f"TOTALS:\t{len(matches)} out of {view.size()} torrents")
                    self.emit(
                        summary.min, item_formatter=lambda i: "MIN:\t" + i.rstrip()
                    )
                    self.emit(
                        summary.average, item_formatter=lambda i: "AVG:\t" + i.rstrip()
                    )
                    self.emit(
                        summary.max, item_formatter=lambda i: "MAX:\t" + i.rstrip()
                    )
                    self.emit(
                        summary.total, item_formatter=lambda i: "SUM:\t" + i.rstrip()
                    )

                self.LOG.info(
                    "Dumped %d out of %d torrents.",
                    len(matches),
                    view.size(),
                )
            else:
                self.LOG.info(
                    "Filtered %d out of %d torrents.",
                    len(matches),
                    view.size(),
                )

            self.LOG.debug("RPC stats: %s", r_engine.rpc)


def run():  # pragma: no cover
    """The entry point."""
    ScriptBase.setup()
    RtorrentControl().run()


if __name__ == "__main__":
    run()



""" Metafile Support.

    Copyright (c) 2009, 2010, 2011 The PyroScope Project <pyroscope.project@gmail.com>
"""


import errno
import fnmatch
import hashlib
import logging
import math
import os
import pprint
import re
import sys
import time
import urllib

from pathlib import Path, PurePath
from typing import Callable, Dict, Generator, List, Optional, Union

import bencode  # typing: ignore

from pyrosimple import config, error
from pyrosimple.util import fmt, pymagic
from pyrosimple.util.parts import Bunch


ALLOWED_ROOT_NAME = re.compile(
    r"^[^/\\.~][^/\\]*$"
)  # cannot be absolute or ~user, and cannot have path parts
ALLOWED_PATH_NAME = re.compile(r"^(?:~\d+)?[^/\\~][^/\\]*$")


PASSKEY_RE = re.compile(r"(?<=[/=])[-_0-9a-zA-Z]{5,64}={0,3}(?=[/&]|$)")


PASSKEY_OK = (
    "announce",
    "TrackerServlet",
)


METAFILE_STD_KEYS = [
    _i.split(".")
    for _i in (
        "announce",
        "announce-list",  # BEP-0012
        "comment",
        "created by",
        "creation date",
        "encoding",
        "info",
        "info.length",
        "info.name",
        "info.piece length",
        "info.pieces",
        "info.private",
        "info.files",
        "info.files.length",
        "info.files.path",
    )
]


def console_progress():
    """Return a progress indicator for consoles if
    stdout is a tty.
    """

    def progress(totalhashed, totalsize):
        "Helper"
        msg = " " * 30
        if totalhashed < totalsize:
            msg = f"{totalhashed * 100.0 / totalsize:5.1f}% complete"
        sys.stdout.write(msg + " \r")
        sys.stdout.flush()

    try:
        return progress if sys.stdout.isatty() else None
    except AttributeError:
        return None


def mask_keys(announce_url: str) -> str:
    """Mask any passkeys (hex sequences) in an announce URL."""
    return PASSKEY_RE.sub(
        lambda m: m.group() if m.group() in PASSKEY_OK else "*" * len(m.group()),
        announce_url,
    )


class MaskingPrettyPrinter(pprint.PrettyPrinter):
    """A PrettyPrinter that masks strings in the object tree."""

    def format(
        self, obj, context, maxlevels, level
    ):  # pylint: disable=arguments-renamed
        """Mask obj if it looks like an URL, then pass it to the super class."""
        if isinstance(obj, str) and "://" in obj:
            obj = mask_keys(obj)
        return pprint.PrettyPrinter.format(self, obj, context, maxlevels, level)


def check_info(info):
    """Validate info dict.

    Raise ValueError if validation fails.
    """
    if not isinstance(info, dict):
        raise ValueError("bad metainfo - not a dictionary")

    pieces = info.get("pieces")
    if not isinstance(pieces, bytes) or len(pieces) % 20 != 0:
        raise ValueError("bad metainfo - bad pieces key")

    piece_size = info.get("piece length")
    if not isinstance(piece_size, int) or piece_size <= 0:
        raise ValueError("bad metainfo - illegal piece length")

    name = info.get("name")
    if not isinstance(name, str):
        raise ValueError(f"bad metainfo - bad name (type is {type(name).__name__!r})")
    if not ALLOWED_ROOT_NAME.match(name):
        raise ValueError(f"name {name} disallowed for security reasons")

    if ("files" in info) == ("length" in info):
        raise ValueError("single/multiple file mix")

    if "length" in info:
        length = info.get("length")
        if not isinstance(length, int) or length < 0:
            raise ValueError("bad metainfo - bad length")
    else:
        files = info.get("files")
        if not isinstance(files, (list, tuple)):
            raise ValueError("bad metainfo - bad file list")

        for item in files:
            if not isinstance(item, dict):
                raise ValueError("bad metainfo - bad file value")

            length = item.get("length")
            if not isinstance(length, int) or length < 0:
                raise ValueError("bad metainfo - bad length")

            path = item.get("path")
            if not isinstance(path, (list, tuple)) or not path:
                raise ValueError("bad metainfo - bad path")

            for part in path:
                if not isinstance(part, str):
                    raise ValueError("bad metainfo - bad path dir")
                if part == "..":
                    raise ValueError(
                        "relative path in %s disallowed for security reasons"
                        % "/".join(path)
                    )
                if part and not ALLOWED_PATH_NAME.match(part):
                    raise ValueError(f"path {part} disallowed for security reasons")

        file_paths = [os.sep.join(item["path"]) for item in files]
        if len(set(file_paths)) != len(file_paths):
            raise ValueError("bad metainfo - duplicate path")

    return info


def check_meta(meta):
    """Validate meta dict.

    Raise ValueError if validation fails.
    """
    if not isinstance(meta, dict):
        raise ValueError("bad metadata - not a dictionary")
    if not isinstance(meta.get("announce"), str):
        raise ValueError("bad announce URL - not a string")
    check_info(meta.get("info"))

    return meta


def clean_meta(meta: Dict, including_info: bool = False, logger=None):
    """Clean meta dict. Optionally log changes using the given logger.

    @param logger: If given, a callable accepting a string message.
    @return: Set of keys removed from C{meta}.
    """
    modified = set()

    for key in list(meta.keys()):
        if [key] not in METAFILE_STD_KEYS:
            if logger:
                logger(f"Removing key {key!r}...")
            del meta[key]
            modified.add(key)

    if including_info:
        for key in list(meta["info"].keys()):
            if ["info", key] not in METAFILE_STD_KEYS:
                if logger:
                    logger(f"Removing key {'info.' + key!r}...")
                del meta["info"][key]
                modified.add("info." + key)

        for idx, entry in enumerate(meta["info"].get("files", [])):
            for key in list(entry.keys()):
                if ["info", "files", key] not in METAFILE_STD_KEYS:
                    if logger:
                        logger("Removing key %r from file #%d..." % (key, idx + 1))
                    del entry[key]
                    modified.add("info.files." + key)

            # Remove crap that certain PHP software puts in paths
            entry["path"] = [i for i in entry["path"] if i]

    return modified


def sanitize(meta, diagnostics=False):
    """Try to fix common problems, especially transcode non-standard string encodings."""
    bad_encodings, bad_fields = set(), set()

    def sane_encoding(field, text):
        "Transcoding helper."
        if isinstance(text, str):
            return text.encode("utf-8")
        for encoding in ("utf-8", meta.get("encoding", None), "cp1252"):
            if encoding:
                try:
                    u8_text = text.decode(encoding).encode("utf-8")
                    if encoding != "utf-8":
                        bad_encodings.add(encoding)
                        bad_fields.add(field)
                    return u8_text
                except UnicodeError:
                    continue
        # Broken beyond anything reasonable
        bad_encodings.add("UNKNOWN/EXOTIC")
        bad_fields.add(field)
        return str(text, "utf-8", "replace").replace("\ufffd", "_").encode("utf-8")

    # Go through all string fields and check them
    for field in ("comment", "created by"):
        if field in meta:
            meta[field] = sane_encoding(field, meta[field])

    meta["info"]["name"] = sane_encoding("info name", meta["info"]["name"])

    for entry in meta["info"].get("files", []):
        entry["path"] = [sane_encoding("file path", i) for i in entry["path"]]

    return (meta, bad_encodings, bad_fields) if diagnostics else meta


def assign_fields(meta, assignments: List[str]):
    """Takes a list of C{key=value} strings and assigns them to the
    given metafile. If you want to set nested keys (e.g. "info.source"),
    you have to use a dot as a separator. For exotic keys *containing*
    a dot, double that dot ("dotted..key").

    Numeric values starting with "+" or "-" are converted to integers.

    If just a key name is given (no '='), the field is removed.
    """
    for assignment in assignments:
        try:
            val: Optional[Union[str, int]]
            if "=" in assignment:
                field, val = assignment.split("=", 1)
            else:
                field, val = assignment, None

            if val is not None and val[0] in "+-" and val[1:].isdigit():
                val = int(val, 10)

            # TODO: Allow numerical indices, and "+" for append
            namespace = meta
            keypath = [
                i.replace("\0", ".") for i in field.replace("..", "\0").split(".")
            ]
            for key in keypath[:-1]:
                # Create missing dicts as we go...
                namespace = namespace.setdefault(key, {})
        except (KeyError, IndexError, TypeError, ValueError) as exc:
            raise error.UserError(f"Bad assignment {assignment!r} ({exc})!") from exc
        else:
            if val is None:
                del namespace[keypath[-1]]
            else:
                namespace[keypath[-1]] = val

    return meta


def add_fast_resume(meta, datapath: str):
    """Add fast resume data to a metafile dict."""
    # Get list of files
    files = meta["info"].get("files", None)
    single = files is None
    if single:
        if os.path.isdir(datapath):
            datapath = os.path.join(datapath, meta["info"]["name"])
        files = [
            Bunch(
                path=[os.path.abspath(datapath)],
                length=meta["info"]["length"],
            )
        ]

    # Prepare resume data
    resume = meta.setdefault("libtorrent_resume", {})
    resume["bitfield"] = len(meta["info"]["pieces"]) // 20
    resume["files"] = []
    piece_length = meta["info"]["piece length"]
    offset = 0

    for fileinfo in files:
        # Get the path into the filesystem
        filepath = Path(*fileinfo["path"])
        if not single:
            filepath = Path(datapath, filepath)

        # Check file size
        if os.path.getsize(filepath) != fileinfo["length"]:
            raise OSError(
                errno.EINVAL,
                "File size mismatch for %r [is %d, expected %d]"
                % (
                    filepath,
                    os.path.getsize(filepath),
                    fileinfo["length"],
                ),
            )

        # Add resume data for this file
        resume["files"].append(
            dict(
                priority=1,
                mtime=int(os.path.getmtime(filepath)),
                completed=(offset + fileinfo["length"] + piece_length - 1)
                // piece_length
                - offset // piece_length,
            )
        )
        offset += fileinfo["length"]

    return meta


def info_hash(metadata) -> str:
    """Return info hash as a string."""
    return hashlib.sha1(bencode.encode(metadata["info"])).hexdigest().upper()


def data_size(metadata) -> int:
    """Calculate the size of a torrent based on parsed metadata."""
    info = metadata["info"]

    if "length" in info:
        # Single file
        return int(info["length"])
    # Directory structure
    return sum(f["length"] for f in info["files"])


def checked_open(filename: str, log: logging.Logger = None):
    """Open and validate the given metafile.
    Optionally provide diagnostics on the passed logger, for
    invalid metafiles, which then just cause a warning but no exception.
    "quiet" can supress that warning.
    """
    with open(filename, "rb") as handle:
        raw_data = handle.read()
    data: Dict = bencode.decode(raw_data)

    try:
        check_meta(data)
        if raw_data != bencode.encode(data):
            raise ValueError("Bad bencoded data - dict keys out of order?")
    except ValueError as exc:
        if log:
            log.warning(f"{filename}: {exc}")
        else:
            raise

    return data


class Metafile:
    """A torrent metafile."""

    # Patterns of names to ignore
    IGNORE_GLOB = [
        "core",
        "CVS",
        ".*",
        "*~",
        "*.swp",
        "*.tmp",
        "*.bak",
        "[Tt]humbs.db",
        "[Dd]esktop.ini",
        "ehthumbs_vista.db",
    ]

    def __init__(self, filename, datapath=None):
        """Initialize metafile."""
        self.filename = filename
        self.progress = None
        self.datapath = datapath
        self._datapath = None
        self.ignore = self.IGNORE_GLOB[:]
        self.LOG = pymagic.get_class_logger(self)

    def _get_datapath(self):
        """Get a valid datapath, else raise an exception."""
        if self._datapath is None:
            raise OSError(
                errno.ENOENT, f"You didn't provide any datapath for {self.filename!r}"
            )

        return self._datapath

    def _set_datapath(self, datapath: str):
        """Set a datapath."""
        if datapath:
            self._datapath = datapath.rstrip(os.sep)
        else:
            self._datapath = None

    datapath = property(_get_datapath, _set_datapath)

    def walk(self) -> Generator[Path, None, None]:
        """Generate paths in "self.datapath"."""
        # Directory?
        if os.path.isdir(self.datapath):
            # Walk the directory tree
            for dirpath, dirnames, filenames in os.walk(
                self.datapath
            ):  # , followlinks=True):
                # Don't scan blacklisted directories
                for bad in dirnames[:]:
                    if any(fnmatch.fnmatch(bad, pattern) for pattern in self.ignore):
                        dirnames.remove(bad)

                # Yield all filenames that aren't blacklisted
                for filename in filenames:
                    if not any(
                        fnmatch.fnmatch(filename, pattern) for pattern in self.ignore
                    ):
                        yield Path(dirpath, filename)

        # Single file
        else:
            # Yield the filename
            yield Path(self.datapath)

    def _calc_size(self) -> int:
        """Get total size of "self.datapath"."""
        return sum(os.path.getsize(filename) for filename in self.walk())

    def _make_info(self, piece_size: int, progress, walker, piece_callback=None):
        """Create info dict."""
        # These collect the file descriptions and piece hashes
        file_list = []
        pieces = []

        # Initialize progress state
        hashing_secs = time.time()
        totalsize = self._calc_size()
        totalhashed = 0

        # Start a new piece
        sha1sum = hashlib.sha1()
        done = 0
        filename = None

        # Hash all files
        for filename in walker:
            # Assemble file info
            filepath = Path(filename)
            filesize = filepath.stat().st_size
            filepath = filepath.relative_to(self.datapath)
            file_list.append(
                {
                    "length": filesize,
                    "path": PurePath(filepath).parts,
                }
            )
            self.LOG.debug("Hashing '%s', size %d...", filename, filesize)

            # Open file and hash it
            fileoffset = 0
            with open(filename, "rb") as handle:
                while fileoffset < filesize:
                    # Read rest of piece or file, whatever is smaller
                    chunk = handle.read(min(filesize - fileoffset, piece_size - done))
                    sha1sum.update(chunk)
                    done += len(chunk)
                    fileoffset += len(chunk)
                    totalhashed += len(chunk)

                    # Piece is done
                    if done == piece_size:
                        pieces.append(sha1sum.digest())
                        if piece_callback:
                            piece_callback(filename, pieces[-1])

                        # Start a new piece
                        sha1sum = hashlib.sha1()
                        done = 0

                    # Report progress
                    if progress:
                        progress(totalhashed, totalsize)

        # Add hash of partial last piece
        if done > 0:
            pieces.append(sha1sum.digest())
            if piece_callback:
                piece_callback(filename, pieces[-1])

        # Build the meta dict
        metainfo = {
            "pieces": b"".join(pieces),
            "piece length": piece_size,
            "name": os.path.basename(self.datapath),
        }

        # Handle directory/FIFO vs. single file
        if os.path.isdir(self.datapath):
            metainfo["files"] = file_list
        else:
            metainfo["length"] = totalhashed

        hashing_secs = time.time() - hashing_secs
        self.LOG.info(
            "Hashing of %s took %.1f secs (%s/s)",
            fmt.human_size(totalhashed).strip(),
            hashing_secs,
            fmt.human_size(totalhashed / hashing_secs).strip(),
        )

        # Return validated info dict
        return check_info(metainfo), totalhashed

    def _make_meta(
        self,
        tracker_url: str,
        root_name: str,
        private: bool,
        progress: Optional[Callable] = None,
        piece_size: int = 0,
    ):
        """Create torrent dict."""
        if piece_size <= 0:
            total_size = self._calc_size()
            if total_size:
                piece_size_exp = int(math.log(total_size) / math.log(2)) - 9
            else:
                piece_size_exp = 0

            piece_size_exp = min(max(15, piece_size_exp), 24)
            piece_size = 2**piece_size_exp

        # Build info hash
        info, totalhashed = self._make_info(piece_size, progress, sorted(self.walk()))

        # Set private flag
        if private:
            info["private"] = 1

        # Freely chosen root name (default is basename of the data path)
        if root_name:
            info["name"] = root_name

        # Torrent metadata
        meta = {
            "info": info,
            "announce": tracker_url.strip(),
        }

        # Return validated meta dict
        return check_meta(meta), totalhashed

    def create(
        self,
        datapath,
        tracker_urls,
        comment=None,
        root_name=None,
        created_by=None,
        private: bool = False,
        no_date: bool = False,
        progress=None,
        callback=None,
    ):
        """Create a metafile with the path given on object creation.
        Returns the last metafile dict that was written (as an object, not bencoded).
        """
        if datapath:
            self.datapath = datapath

        try:
            tracker_urls = ["" + tracker_urls]
        except TypeError:
            tracker_urls = list(tracker_urls)
        multi_mode = len(tracker_urls) > 1

        # TODO add optimization so the hashing happens only once for multiple URLs!
        for tracker_url in tracker_urls:
            # Lookup announce URLs from config file
            try:
                if urllib.parse.urlparse(tracker_url).scheme:
                    tracker_alias = (
                        urllib.parse.urlparse(tracker_url)
                        .netloc.split(":")[0]
                        .split(".")
                    )
                    tracker_alias = tracker_alias[-2 if len(tracker_alias) > 1 else 0]
                else:
                    tracker_alias, tracker_url = config.lookup_announce_url(tracker_url)
                    tracker_url = tracker_url[0]
            except (KeyError, IndexError) as exc:
                raise error.UserError(
                    f"Bad tracker URL {tracker_url!r}, or unknown alias!"
                ) from exc

            # Determine metafile name
            output_name = self.filename
            if multi_mode:
                # Add 2nd level of announce URL domain to metafile name
                output_name = list(os.path.splitext(output_name))
                try:
                    output_name[1:1] = "-" + tracker_alias
                except (IndexError,):
                    self.LOG.error("Malformed announce URL %r, skipping!", tracker_url)
                    continue
                output_name = "".join(output_name)

            # Hash the data
            self.LOG.info(
                "Creating %r from %r...",
                output_name,
                self.datapath,
            )
            meta, _ = self._make_meta(tracker_url, root_name, private, progress)

            # Add optional fields
            if comment:
                meta["comment"] = comment
            if created_by:
                meta["created by"] = created_by
            if not no_date:
                meta["creation date"] = int(time.time())
            if callback:
                callback(meta)

            # Write metafile to disk
            self.LOG.debug("Writing %r...", output_name)
            with open(output_name, "wb") as fh:
                fh.write(bencode.encode(meta))

        return meta

    def check(self, metainfo, datapath, progress=None):
        """Check piece hashes of a metafile against the given datapath."""
        if datapath:
            self.datapath = datapath

        def check_piece(filename, piece):
            "Callback for new piece"
            if (
                piece
                != metainfo["info"]["pieces"][
                    check_piece.piece_index : check_piece.piece_index + 20
                ]
            ):
                self.LOG.warning(
                    "Piece #%d: Hashes differ in file %r",
                    check_piece.piece_index // 20,
                    filename,
                )
            check_piece.piece_index += 20

        check_piece.piece_index = 0

        datameta, _ = self._make_info(
            int(metainfo["info"]["piece length"]),
            progress,
            [datapath]
            if "length" in metainfo["info"]
            else (
                os.path.join(*([datapath] + i["path"]))
                for i in metainfo["info"]["files"]
            ),
            piece_callback=check_piece,
        )
        return datameta["pieces"] == metainfo["info"]["pieces"]

    def listing(self, masked=True):
        """List torrent info & contents. Returns a list of formatted lines."""
        # Assemble data
        with open(self.filename, "rb") as fh:
            metainfo = bencode.decode(fh.read())
        bad_encodings = []
        bad_fields = []
        announce = metainfo["announce"]
        info = metainfo["info"]
        infohash = hashlib.sha1(bencode.encode(info))

        total_size = data_size(metainfo)
        piece_length = info["piece length"]
        piece_number, last_piece_length = divmod(total_size, piece_length)

        # Build result
        result: List[str] = [
            f"NAME {Path(self.filename).name}",
            "SIZE %s (%i * %s + %s)"
            % (
                fmt.human_size(total_size).strip(),
                piece_number,
                fmt.human_size(piece_length).strip(),
                fmt.human_size(last_piece_length).strip(),
            ),
            "META %s (pieces %s %.1f%%)"
            % (
                fmt.human_size(os.path.getsize(self.filename)).strip(),
                fmt.human_size(len(info["pieces"])).strip(),
                100.0 * len(info["pieces"]) / os.path.getsize(self.filename),
            ),
            f"HASH {infohash.hexdigest().upper()}",
            f"URL  {(mask_keys if masked else str)(announce)}",
            "PRV  %s"
            % (
                "YES (DHT/PEX disabled)"
                if info.get("private")
                else "NO (DHT/PEX enabled)"
            ),
            "TIME %s"
            % (
                "N/A"
                if "creation date" not in metainfo
                else time.strftime(
                    "%Y-%m-%d %H:%M:%S", time.localtime(metainfo["creation date"])
                )
            ),
        ]

        for label, key in (("BY  ", "created by"), ("REM ", "comment")):
            if key in metainfo:
                result.append(f"{label} {metainfo.get(key, 'N/A')}")

        result.extend(
            [
                "",
                "FILE LISTING%s"
                % ("" if "length" in info else " [%d file(s)]" % len(info["files"]),),
            ]
        )
        if "length" in info:
            # Single file
            result.append(
                "%-69s%9s"
                % (
                    info["name"],
                    fmt.human_size(total_size),
                )
            )
        else:
            # Directory structure
            result.append(f"{info['name']}/")
            oldpaths = [None] * 99
            for entry in info["files"]:
                # Remove crap that certain PHP software puts in paths
                entry_path = [i for i in entry["path"] if i]

                for idx, item in enumerate(entry_path[:-1]):
                    if item != oldpaths[idx]:
                        result.append(f"{' ' * (4 * (idx + 1))}{item}/")
                        oldpaths[idx] = item
                result.append(
                    "%-69s%9s"
                    % (
                        " " * (4 * len(entry_path)) + entry_path[-1],
                        fmt.human_size(entry["length"]),
                    )
                )

        if bad_encodings:
            result.extend(
                [
                    "",
                    "WARNING: Bad encoding(s) {} in these fields: {}".format(
                        ", ".join(sorted(bad_encodings)), ", ".join(sorted(bad_fields))
                    ),
                    "Use the --raw option to inspect these encoding issues.",
                ]
            )

        return result



""" Statistics data.

    Copyright (c) 2014 The PyroScope Project <pyroscope.project@gmail.com>
"""


import time


def engine_data(engine):
    """Get important performance data and metadata from rTorrent."""
    views = (
        "default",
        "main",
        "started",
        "stopped",
        "complete",
        "incomplete",
        "seeding",
        "leeching",
        "active",
        "messages",
    )
    methods = [
        "throttle.global_up.rate",
        "throttle.global_up.max_rate",
        "throttle.global_down.rate",
        "throttle.global_down.max_rate",
    ]

    # Get data via multicall
    proxy = engine.open()
    calls = [dict(methodName=method, params=[]) for method in methods] + [
        dict(methodName="view.size", params=["", view]) for view in views
    ]
    result = proxy.system.multicall(calls, flatten=True)

    # Build result object
    data = dict(
        now=time.time(),
        engine_id=engine.engine_id,
        versions=engine.versions,
        uptime=engine.uptime,
        upload=[result[0], result[1]],
        download=[result[2], result[3]],
        views={name: result[4 + i] for i, name in enumerate(views)},
    )

    return data



""" Category management.

    Copyright (c) 2017 The PyroScope Project <pyroscope.project@gmail.com>
"""


from pyrosimple import error
from pyrosimple.scripts.base import ScriptBase, ScriptBaseWithConfig
from pyrosimple.torrent import rtorrent
from pyrosimple.util import rpc


class CategoryManager(ScriptBaseWithConfig):
    """Rotate through category views."""

    PREFIX = "category_"
    PREFIX_LEN = len(PREFIX)

    # argument description for the usage information
    ARGS_HELP = ""

    def add_options(self):
        """Add program options."""
        super().add_options()

        self.add_bool_option("-l", "--list", help="list added category views")
        self.add_bool_option("-n", "--next", help="switch to next category view")
        self.add_bool_option("-p", "--prev", help="switch to previous category view")
        self.add_bool_option(
            "-u", "--update", help="filter the current category view again"
        )

    def mainloop(self):
        """Manage category views."""
        # Get client state
        proxy = rtorrent.RtorrentEngine().open()
        views = [x for x in sorted(proxy.view.list()) if x.startswith(self.PREFIX)]

        current_view = real_current_view = proxy.ui.current_view()
        if current_view not in views:
            if views:
                current_view = views[0]
            else:
                raise error.UserError(
                    f"There are no '{self.PREFIX}*' views defined at all!"
                )

        # Check options
        if self.options.list:
            for name in sorted(views):
                print(
                    "{} {:5d} {}".format(
                        "*" if name == real_current_view else " ",
                        proxy.view.size(rpc.NOHASH, name),
                        name[self.PREFIX_LEN :],
                    )
                )

        elif self.options.next or self.options.prev or self.options.update:
            # Determine next in line
            if self.options.update:
                new_view = current_view
            else:
                new_view = (views * 2)[
                    views.index(current_view) + (1 if self.options.next else -1)
                ]

            self.LOG.info(
                "{} category view '{}'.".format(
                    "Updating" if self.options.update else "Switching to", new_view
                )
            )

            # Update and switch to filtered view
            proxy.pyro.category.update(rpc.NOHASH, new_view[self.PREFIX_LEN :])
            proxy.ui.current_view.set(new_view)

        else:
            self.LOG.info(
                "Current category view is '%s'.", current_view[self.PREFIX_LEN :]
            )
            self.LOG.info("Use '--help' to get usage information.")


def run():  # pragma: no cover
    """The entry point."""
    ScriptBase.setup()
    CategoryManager().run()


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pyrosimple',
 'pyrosimple.daemon',
 'pyrosimple.data.config',
 'pyrosimple.io',
 'pyrosimple.job',
 'pyrosimple.scripts',
 'pyrosimple.torrent',
 'pyrosimple.ui',
 'pyrosimple.util']

package_data = \
{'': ['*'],
 'pyrosimple': ['data/htdocs/*',
                'data/htdocs/css/*',
                'data/htdocs/img/*',
                'data/htdocs/js/*',
                'data/img/*',
                'data/screenlet/*',
                'data/screenlet/themes/blueish/*',
                'data/screenlet/themes/default/*'],
 'pyrosimple.data.config': ['color-schemes/*',
                            'rtorrent.d/*',
                            'templates/*',
                            'templates/conky/*']}

install_requires = \
['Jinja2>=3.1.0,<4.0.0',
 'bencode.py>=4.0.0,<5.0.0',
 'dynaconf[toml]>=3.1.0,<4.0.0',
 'parsimonious>=0.9.0,<0.10.0',
 'prometheus-client>=0.14.1,<0.15.0',
 'python-daemon>=2.3.0,<3.0.0']

extras_require = \
{':python_version < "3.9"': ['importlib-resources>=5.4.0,<6.0.0'],
 'torque': ['APScheduler>=3.9.0,<4.0.0', 'pyinotify>=0.9.6,<0.10.0']}

entry_points = \
{'console_scripts': ['chtor = pyrosimple.scripts.chtor:run',
                     'lstor = pyrosimple.scripts.lstor:run',
                     'mktor = pyrosimple.scripts.mktor:run',
                     'pyroadmin = pyrosimple.scripts.pyroadmin:run',
                     'pyrotorque = pyrosimple.scripts.pyrotorque:run',
                     'rtcontrol = pyrosimple.scripts.rtcontrol:run',
                     'rtxmlrpc = pyrosimple.scripts.rtxmlrpc:run']}

setup_kwargs = {
    'name': 'pyrosimple',
    'version': '2.0.0rc1',
    'description': '',
    'long_description': "# pyrosimple\n\n[![GitHub Workflow Status](https://img.shields.io/github/workflow/status/kannibalox/pyrosimple/Pylint)](https://github.com/kannibalox/pyrosimple/actions/workflows/pylint.yml)\n[![PyPI](https://img.shields.io/pypi/v/pyrosimple)](https://pypi.org/project/pyrosimple/)\n![PyPI -\tPython Version](https://img.shields.io/pypi/pyversions/pyrosimple)\n\nA overhauled Python 3 fork  of the [pyrocore tools](https://github.com/pyroscope/pyrocore), for working with the [rTorrent client](https://github.com/rakshasa/rtorrent).\n\n## Installation\n\n```bash\npip install pyrosimple\n```\n\nSee the [documentation for usage](https://kannibalox.github.io/pyrosimple/). If you've used rtcontrol/rtxmlrpc before, you should feel right at home.\n\n## What's the point of this?\n\nThe pyrocore tools are great, but being stuck on python 2, along with the complicated install procedure made integrating both the tools and the code into other processes very painful.\n\n## Significant differences from pyrocore\n\nThe following lists are not exhaustive, and don't cover many of the internal improvements and refactoring.\n\n- Only supports python 3 and rtorrent 0.9.8+ (0.9.6/0.9.7 should still work just fine, but aren't officially supported)\n- Simpler poetry-based build/install system\n- Everything in one package (no separate pyrobase)\n  - Use external lib for bencode\n\n### Added\n- Multi-instance support for rtcontrol/rtxmlrpc\n- Replaced Tempita with jinja2\n- Support for JSON-RPC (only implemented by https://github.com/jesec/rtorrent)\n- pyrotorque job to move torrents between hosts\n- pyrotorque job to move torrent paths\n\n### Removed\n- the `rtsweep`, `rtmv`, and `rtevent` commands\n- `pyrotorque`'s guard file, influxdb job and web server\n\n### Changed\n- `rtxmlrpc`'s `--raw` now outputs JSON\n",
    'author': 'kannibalox',
    'author_email': 'kannibalox@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kannibalox/pyrosimple',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4',
}


setup(**setup_kwargs)
